#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the Spark Step Operator."""

import subprocess
from typing import TYPE_CHECKING, Dict, List, Optional, Type, cast

from pyspark.conf import SparkConf

from zenml.client import Client
from zenml.integrations.spark.flavors.spark_step_operator_flavor import (
    SparkStepOperatorConfig,
    SparkStepOperatorSettings,
)
from zenml.integrations.spark.step_operators.spark_entrypoint_configuration import (
    SparkEntrypointConfiguration,
)
from zenml.logger import get_logger
from zenml.step_operators import BaseStepOperator

logger = get_logger(__name__)
if TYPE_CHECKING:
    from zenml.config import ResourceSettings
    from zenml.config.base_settings import BaseSettings
    from zenml.config.step_run_info import StepRunInfo


class SparkStepOperator(BaseStepOperator):
    """Base class for all Spark-related step operators."""

    @property
    def config(self) -> SparkStepOperatorConfig:
        """Returns the `SparkStepOperatorConfig` config.

        Returns:
            The configuration.
        """
        return cast(SparkStepOperatorConfig, self._config)

    @property
    def settings_class(self) -> Optional[Type["BaseSettings"]]:
        """Settings class for the Spark step operator.

        Returns:
            The settings class.
        """
        return SparkStepOperatorSettings

    @property
    def application_path(self) -> Optional[str]:
        """Optional method for providing the application path.

        This is especially critical when using 'spark-submit' as it defines the
        path (to the application in the environment where Spark is running)
        which is used within the command.

        For more information on how to set this property please check:

        https://spark.apache.org/docs/latest/submitting-applications.html#advanced-dependency-management

        Returns:
            The path to the application entrypoint
        """
        return None

    def _resource_configuration(
        self,
        spark_config: SparkConf,
        resource_settings: "ResourceSettings",
    ) -> None:
        """Configures Spark to handle the resource settings.

        This should serve as the layer between our ResourceSettings
        and Spark's own ways of configuring its resources.

        Note: This is still work-in-progress. In the future, we would like to
        enable much more than executor cores and memory with a dedicated
        ResourceSettings object.

        Args:
            spark_config: a SparkConf object which collects all the
                configuration parameters
            resource_settings: the resource settings for this step
        """
        if resource_settings.cpu_count:
            spark_config.set(
                "spark.executor.cores",
                str(int(resource_settings.cpu_count)),
            )

        if resource_settings.memory:
            # TODO[LOW]: Fix the conversion of the memory unit with a new
            #   type of resource configuration.
            spark_config.set(
                "spark.executor.memory",
                resource_settings.memory.lower().strip("b"),
            )

    def _backend_configuration(
        self,
        spark_config: SparkConf,
        info: "StepRunInfo",
        environment: Dict[str, str],
    ) -> None:
        """Configures Spark to handle backends like YARN, Mesos or Kubernetes.

        Args:
            spark_config: a SparkConf object which collects all the
                configuration parameters
            info: Information about the step run.
            environment: Environment variables to set.
        """

    def _io_configuration(self, spark_config: SparkConf) -> None:
        """Configures Spark to handle different input/output sources.

        When you work with the Spark integration, you get materializers
        such as SparkDataFrameMaterializer, SparkModelMaterializer. However, in
        many cases, these materializer work only if the environment, where
        Spark is running, is configured according to the artifact store.

        Take s3 as an example. When you want to save a dataframe to an S3
        artifact store, you need to provide configuration parameters such as,
        '"spark.hadoop.fs.s3.impl=org.apache.hadoop.fs.s3a.S3AFileSystem" to
        Spark. This method aims to provide these configuration parameters.

        Args:
            spark_config: a SparkConf object which collects all the
                configuration parameters

        Raises:
            RuntimeError: when the step operator is being used with an S3
                artifact store and the artifact store does not have the
                required authentication
        """
        # Get active artifact store
        client = Client()
        artifact_store = client.active_stack.artifact_store

        from zenml.integrations.s3 import S3_ARTIFACT_STORE_FLAVOR

        # If S3, preconfigure the spark session
        if artifact_store.flavor == S3_ARTIFACT_STORE_FLAVOR:
            (
                key,
                secret,
                _,
            ) = artifact_store._get_credentials()  # type:ignore[attr-defined]
            if key and secret:
                spark_config.setAll(
                    [
                        ("spark.hadoop.fs.s3a.fast.upload", "true"),
                        (
                            "spark.hadoop.fs.s3.impl",
                            "org.apache.hadoop.fs.s3a.S3AFileSystem",
                        ),
                        (
                            "spark.hadoop.fs.AbstractFileSystem.s3.impl",
                            "org.apache.hadoop.fs.s3a.S3A",
                        ),
                        (
                            "spark.hadoop.fs.s3a.aws.credentials.provider",
                            "com.amazonaws.auth.DefaultAWSCredentialsProviderChain",
                        ),
                        ("spark.hadoop.fs.s3a.access.key", f"{key}"),
                        ("spark.hadoop.fs.s3a.secret.key", f"{secret}"),
                    ]
                )
            else:
                raise RuntimeError(
                    "When you use an Spark step operator with an S3 artifact "
                    "store, please make sure that your artifact store has"
                    "defined the required credentials namely the access key "
                    "and the secret access key."
                )
        else:
            logger.warning(
                "In most cases, the Spark step operator requires additional "
                "configuration based on the artifact store flavor you are "
                "using. That also means, that when you use this step operator "
                "with certain artifact store flavor, ZenML can take care of "
                "the pre-configuration. However, the artifact store flavor "
                f"'{artifact_store.flavor}' featured in this stack is not "
                f"known to this step operator and it might require additional "
                f"configuration."
            )

    def _additional_configuration(
        self, spark_config: SparkConf, settings: SparkStepOperatorSettings
    ) -> None:
        """Appends the user-defined configuration parameters.

        Args:
            spark_config: a SparkConf object which collects all the
                configuration parameters
            settings: Step operator settings for the current step run.
        """
        # Add the additional parameters
        if settings.submit_kwargs:
            for k, v in settings.submit_kwargs.items():
                spark_config.set(k, v)

    def _launch_spark_job(
        self,
        spark_config: SparkConf,
        deploy_mode: str,
        entrypoint_command: List[str],
    ) -> None:
        """Generates and executes a spark-submit command.

        Args:
            spark_config: a SparkConf object which collects all the
                configuration parameters
            deploy_mode: The spark deploy mode to use.
            entrypoint_command: The entrypoint command to run.

        Raises:
            RuntimeError: if the spark-submit fails
        """
        # Base spark-submit command
        command = [
            f"spark-submit "
            f"--master {self.config.master} "
            f"--deploy-mode {deploy_mode}"
        ]

        # Add the configuration parameters
        command += [f"--conf {c[0]}={c[1]}" for c in spark_config.getAll()]

        # Add the application path
        command.append(self.application_path)  # type: ignore[arg-type]

        # Update the default step operator command to use the spark entrypoint
        # configuration
        original_args = SparkEntrypointConfiguration._parse_arguments(
            entrypoint_command
        )
        command += SparkEntrypointConfiguration.get_entrypoint_arguments(
            **original_args
        )

        final_command = " ".join(command)

        # Execute the spark-submit
        process = subprocess.Popen(
            final_command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            shell=True,  # nosec
        )
        stdout, stderr = process.communicate()

        if process.returncode != 0:
            raise RuntimeError(stderr)
        print(stdout)

    def launch(
        self,
        info: "StepRunInfo",
        entrypoint_command: List[str],
        environment: Dict[str, str],
    ) -> None:
        """Launches a step on Spark.

        Args:
            info: Information about the step run.
            entrypoint_command: Command that executes the step.
            environment: Environment variables to set in the step operator
                environment.
        """
        settings = cast(SparkStepOperatorSettings, self.get_settings(info))
        # Start off with an empty configuration
        conf = SparkConf()

        # Add the resource configuration such as cores, memory.
        self._resource_configuration(
            spark_config=conf,
            resource_settings=info.config.resource_settings,
        )

        # Add the backend configuration such as namespace, docker images names.
        self._backend_configuration(
            spark_config=conf, info=info, environment=environment
        )

        # Add the IO configuration for the inputs and the outputs
        self._io_configuration(
            spark_config=conf,
        )

        # Add any additional configuration given by the user.
        self._additional_configuration(spark_config=conf, settings=settings)

        info.force_write_logs()

        # Generate a spark-submit command given the configuration
        self._launch_spark_job(
            spark_config=conf,
            deploy_mode=settings.deploy_mode,
            entrypoint_command=entrypoint_command,
        )



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""SQLModel implementation of user tables."""

import json
from typing import TYPE_CHECKING, Any, List, Optional, Union
from uuid import UUID

from sqlalchemy import TEXT, Column, UniqueConstraint
from sqlmodel import Field, Relationship

from zenml.models import (
    ServiceAccountInternalRequest,
    ServiceAccountRequest,
    ServiceAccountResponse,
    ServiceAccountResponseBody,
    ServiceAccountResponseMetadata,
    ServiceAccountUpdate,
    UserRequest,
    UserResponse,
    UserResponseBody,
    UserResponseMetadata,
    UserUpdate,
)
from zenml.utils.time_utils import utc_now
from zenml.zen_stores.schemas.base_schemas import NamedSchema
from zenml.zen_stores.schemas.schema_utils import build_foreign_key_field

if TYPE_CHECKING:
    from zenml.zen_stores.schemas import (
        APIKeySchema,
        ArtifactSchema,
        ArtifactVersionSchema,
        CodeRepositorySchema,
        DeploymentSchema,
        FlavorSchema,
        ModelSchema,
        ModelVersionSchema,
        OAuthDeviceSchema,
        PipelineBuildSchema,
        PipelineRunSchema,
        PipelineSchema,
        PipelineSnapshotSchema,
        RunMetadataSchema,
        RunTemplateSchema,
        ScheduleSchema,
        SecretSchema,
        ServiceConnectorSchema,
        ServiceSchema,
        StackComponentSchema,
        StackSchema,
        StepRunSchema,
        TagSchema,
        TriggerSchema,
    )


class UserSchema(NamedSchema, table=True):
    """SQL Model for users."""

    __tablename__ = "user"
    __table_args__ = (UniqueConstraint("name", "is_service_account"),)

    is_service_account: bool = Field(default=False)
    full_name: str
    description: Optional[str] = Field(sa_column=Column(TEXT, nullable=True))
    email: Optional[str] = Field(nullable=True)
    avatar_url: Optional[str] = Field(
        default=None,
        sa_column=Column(TEXT, nullable=True),
    )
    active: bool
    password: Optional[str] = Field(nullable=True)
    activation_token: Optional[str] = Field(nullable=True)
    email_opted_in: Optional[bool] = Field(nullable=True)
    external_user_id: Optional[UUID] = Field(nullable=True)
    is_admin: bool = Field(default=False)
    user_metadata: Optional[str] = Field(nullable=True)

    default_project_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target="project",
        source_column="default_project_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )

    stacks: List["StackSchema"] = Relationship(back_populates="user")
    components: List["StackComponentSchema"] = Relationship(
        back_populates="user",
    )
    flavors: List["FlavorSchema"] = Relationship(back_populates="user")
    pipelines: List["PipelineSchema"] = Relationship(back_populates="user")
    schedules: List["ScheduleSchema"] = Relationship(
        back_populates="user",
    )
    runs: List["PipelineRunSchema"] = Relationship(back_populates="user")
    run_templates: List["RunTemplateSchema"] = Relationship(
        back_populates="user",
    )
    step_runs: List["StepRunSchema"] = Relationship(back_populates="user")
    builds: List["PipelineBuildSchema"] = Relationship(back_populates="user")
    artifacts: List["ArtifactSchema"] = Relationship(back_populates="user")
    artifact_versions: List["ArtifactVersionSchema"] = Relationship(
        back_populates="user"
    )
    run_metadata: List["RunMetadataSchema"] = Relationship(
        back_populates="user"
    )
    secrets: List["SecretSchema"] = Relationship(
        back_populates="user",
        sa_relationship_kwargs={"cascade": "delete"},
    )
    snapshots: List["PipelineSnapshotSchema"] = Relationship(
        back_populates="user",
    )
    code_repositories: List["CodeRepositorySchema"] = Relationship(
        back_populates="user",
    )
    services: List["ServiceSchema"] = Relationship(back_populates="user")
    service_connectors: List["ServiceConnectorSchema"] = Relationship(
        back_populates="user",
    )
    models: List["ModelSchema"] = Relationship(
        back_populates="user",
    )
    model_versions: List["ModelVersionSchema"] = Relationship(
        back_populates="user",
    )
    auth_devices: List["OAuthDeviceSchema"] = Relationship(
        back_populates="user",
        sa_relationship_kwargs={"cascade": "delete"},
    )
    api_keys: List["APIKeySchema"] = Relationship(
        back_populates="service_account",
        sa_relationship_kwargs={"cascade": "delete"},
    )
    deployments: List["DeploymentSchema"] = Relationship(
        back_populates="user",
    )
    tags: List["TagSchema"] = Relationship(
        back_populates="user",
    )
    triggers: List["TriggerSchema"] = Relationship(
        back_populates="user",
    )

    @classmethod
    def from_user_request(cls, model: UserRequest) -> "UserSchema":
        """Create a `UserSchema` from a `UserRequest`.

        Args:
            model: The `UserRequest` from which to create the schema.

        Returns:
            The created `UserSchema`.
        """
        return cls(
            name=model.name,
            full_name=model.full_name,
            avatar_url=model.avatar_url,
            active=model.active,
            password=model.create_hashed_password(),
            activation_token=model.create_hashed_activation_token(),
            external_user_id=model.external_user_id,
            email_opted_in=model.email_opted_in,
            email=model.email,
            is_service_account=False,
            is_admin=model.is_admin,
            user_metadata=json.dumps(model.user_metadata)
            if model.user_metadata
            else None,
        )

    @classmethod
    def from_service_account_request(
        cls, model: Union[ServiceAccountRequest, ServiceAccountInternalRequest]
    ) -> "UserSchema":
        """Create a `UserSchema` from a Service Account request.

        Args:
            model: The `ServiceAccountRequest` or `ServiceAccountInternalRequest`
                from which to create the schema.

        Returns:
            The created `UserSchema`.
        """
        return cls(
            name=model.name,
            full_name=model.full_name,
            description=model.description or "",
            external_user_id=model.external_user_id
            if isinstance(model, ServiceAccountInternalRequest)
            else None,
            active=model.active,
            is_service_account=True,
            email_opted_in=False,
            is_admin=False,
            avatar_url=model.avatar_url,
        )

    def update_user(self, user_update: UserUpdate) -> "UserSchema":
        """Update a `UserSchema` from a `UserUpdate`.

        Args:
            user_update: The `UserUpdate` from which to update the schema.

        Returns:
            The updated `UserSchema`.
        """
        for field, value in user_update.model_dump(exclude_unset=True).items():
            if field == "old_password":
                continue

            if field == "password":
                setattr(self, field, user_update.create_hashed_password())
            elif field == "activation_token":
                setattr(
                    self, field, user_update.create_hashed_activation_token()
                )
            elif field == "user_metadata":
                if value is not None:
                    self.user_metadata = json.dumps(value)
            else:
                setattr(self, field, value)

        self.updated = utc_now()
        return self

    def update_service_account(
        self, service_account_update: ServiceAccountUpdate
    ) -> "UserSchema":
        """Update a `UserSchema` from a `ServiceAccountUpdate`.

        Args:
            service_account_update: The `ServiceAccountUpdate` from which
                to update the schema.

        Returns:
            The updated `UserSchema`.
        """
        for field, value in service_account_update.model_dump(
            exclude_none=True
        ).items():
            setattr(self, field, value)

        self.updated = utc_now()
        return self

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        include_private: bool = False,
        **kwargs: Any,
    ) -> UserResponse:
        """Convert a `UserSchema` to a `UserResponse`.

        Args:
            include_metadata: Whether the metadata will be filled.
            include_resources: Whether the resources will be filled.
            **kwargs: Keyword arguments to allow schema specific logic
            include_private: Whether to include the user private information
                this is to limit the amount of data one can get about other
                users.

        Returns:
            The converted `UserResponse`.
        """
        metadata = None
        if include_metadata:
            metadata = UserResponseMetadata(
                email=self.email if include_private else None,
                external_user_id=self.external_user_id,
                user_metadata=json.loads(self.user_metadata)
                if self.user_metadata
                else {},
            )

        return UserResponse(
            id=self.id,
            name=self.name,
            body=UserResponseBody(
                active=self.active,
                full_name=self.full_name,
                email_opted_in=self.email_opted_in,
                is_service_account=self.is_service_account,
                created=self.created,
                updated=self.updated,
                is_admin=self.is_admin,
                default_project_id=self.default_project_id,
                avatar_url=self.avatar_url,
            ),
            metadata=metadata,
        )

    def to_service_account_model(
        self, include_metadata: bool = False, include_resources: bool = False
    ) -> ServiceAccountResponse:
        """Convert a `UserSchema` to a `ServiceAccountResponse`.

        Args:
             include_metadata: Whether the metadata will be filled.
             include_resources: Whether the resources will be filled.

        Returns:
             The converted `ServiceAccountResponse`.
        """
        metadata = None
        if include_metadata:
            metadata = ServiceAccountResponseMetadata(
                description=self.description or "",
                external_user_id=self.external_user_id,
            )

        body = ServiceAccountResponseBody(
            full_name=self.full_name,
            created=self.created,
            updated=self.updated,
            active=self.active,
            avatar_url=self.avatar_url,
        )

        return ServiceAccountResponse(
            id=self.id,
            name=self.name,
            body=body,
            metadata=metadata,
        )



"""Artifact Version Table [a91762e6be36].

Revision ID: a91762e6be36
Revises: 0.50.0
Create Date: 2023-11-25 11:01:09.217299

"""

from uuid import uuid4

import sqlalchemy as sa
import sqlmodel
from alembic import op

from zenml.utils.time_utils import utc_now

# revision identifiers, used by Alembic.
revision = "a91762e6be36"
down_revision = "0.50.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # Rename old artifact table to artifact_version
    op.rename_table("artifact", "artifact_version")

    # Create new artifact table
    op.create_table(
        "artifact",
        sa.Column("id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("updated", sa.DateTime(), nullable=False),
        sa.Column("name", sa.VARCHAR(length=255), nullable=False),
        sa.Column("has_custom_name", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )

    # Add foreign key column to artifact_version, for now as nullable
    with op.batch_alter_table("artifact_version", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "artifact_id", sqlmodel.sql.sqltypes.GUID(), nullable=True
            )
        )

    # Find all unique artifact names and create a new artifact entry for each
    conn = op.get_bind()
    meta = sa.MetaData()
    meta.reflect(only=("artifact", "artifact_version"), bind=op.get_bind())
    artifacts = sa.Table("artifact", meta)
    artifact_versions = sa.Table("artifact_version", meta)
    artifact_names = conn.execute(
        sa.select(
            artifact_versions.c.name, artifact_versions.c.has_custom_name
        ).distinct()
    ).all()
    unique_artifact_names = {
        name: has_custom_name for name, has_custom_name in artifact_names
    }
    now = utc_now()
    for name, has_custom_name in unique_artifact_names.items():
        conn.execute(
            artifacts.insert().values(
                id=uuid4().hex,
                created=now,
                updated=now,
                name=name,
                has_custom_name=has_custom_name,
            )
        )

    # Set artifact_id column in artifact_version
    conn.execute(
        artifact_versions.update().values(
            artifact_id=sa.select(artifacts.c.id)
            .where(artifacts.c.name == artifact_versions.c.name)
            .scalar_subquery()
        )
    )

    # Make foreign key column non-nullable and create constraint
    with op.batch_alter_table("artifact_version", schema=None) as batch_op:
        batch_op.alter_column(
            "artifact_id", nullable=False, existing_type=sa.CHAR(length=32)
        )
        batch_op.create_foreign_key(
            "fk_artifact_version_artifact_id_artifact",
            "artifact",
            ["artifact_id"],
            ["id"],
            ondelete="CASCADE",
        )

    # Drop old columns from artifact_version
    with op.batch_alter_table("artifact_version", schema=None) as batch_op:
        batch_op.drop_column("name")
        batch_op.drop_column("has_custom_name")

    # Rename constraints `fk_artifact_...` to `fk_artifact_version_...`
    with op.batch_alter_table("artifact_version", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_artifact_artifact_store_id_stack_component", type_="foreignkey"
        )
        batch_op.drop_constraint(
            "fk_artifact_user_id_user", type_="foreignkey"
        )
        batch_op.drop_constraint(
            "fk_artifact_workspace_id_workspace", type_="foreignkey"
        )
    with op.batch_alter_table("artifact_version", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "fk_artifact_version_artifact_store_id_stack_component",
            "stack_component",
            ["artifact_store_id"],
            ["id"],
            ondelete="SET NULL",
        )
        batch_op.create_foreign_key(
            "fk_artifact_version_user_id_user",
            "user",
            ["user_id"],
            ["id"],
            ondelete="SET NULL",
        )
        batch_op.create_foreign_key(
            "fk_artifact_version_workspace_id_workspace",
            "workspace",
            ["workspace_id"],
            ["id"],
            ondelete="CASCADE",
        )

    # Rename constraints of other tables
    with op.batch_alter_table(
        "artifact_visualization", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            "fk_artifact_visualization_artifact_id_artifact",
            type_="foreignkey",
        )
        batch_op.alter_column(
            "artifact_id",
            new_column_name="artifact_version_id",
            existing_type=sa.CHAR(length=32),
        )
    with op.batch_alter_table(
        "artifact_visualization", schema=None
    ) as batch_op:
        batch_op.create_foreign_key(
            "fk_artifact_visualization_artifact_version_id_artifact_version",
            "artifact_version",
            ["artifact_version_id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table(
        "model_versions_artifacts", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            "fk_model_versions_artifacts_artifact_id_artifact",
            type_="foreignkey",
        )
        batch_op.alter_column(
            "artifact_id",
            new_column_name="artifact_version_id",
            existing_type=sa.CHAR(length=32),
        )
    with op.batch_alter_table(
        "model_versions_artifacts", schema=None
    ) as batch_op:
        batch_op.create_foreign_key(
            "fk_model_versions_artifacts_artifact_version_id_artifact_version",
            "artifact_version",
            ["artifact_version_id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table("run_metadata", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_run_metadata_artifact_id_artifact", type_="foreignkey"
        )
        batch_op.alter_column(
            "artifact_id",
            new_column_name="artifact_version_id",
            existing_type=sa.CHAR(length=32),
        )
    with op.batch_alter_table("run_metadata", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "fk_run_metadata_artifact_version_id_artifact_version",
            "artifact_version",
            ["artifact_version_id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table(
        "step_run_input_artifact", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            "fk_step_run_input_artifact_artifact_id_artifact",
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            "fk_step_run_input_artifact_artifact_id_artifact_version",
            "artifact_version",
            ["artifact_id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table(
        "step_run_output_artifact", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            "fk_step_run_output_artifact_artifact_id_artifact",
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            "fk_step_run_output_artifact_artifact_id_artifact_version",
            "artifact_version",
            ["artifact_id"],
            ["id"],
            ondelete="CASCADE",
        )


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision.

    Raises:
        NotImplementedError: Downgrade is not supported for this migration.
    """
    raise NotImplementedError("Downgrade is not supported for this migration.")



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the ZenML Docker deployer."""

import copy
import os
import sys
from typing import (
    Any,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Type,
    cast,
)

import docker.errors as docker_errors
from docker.client import DockerClient
from docker.models.containers import Container
from pydantic import BaseModel

from zenml.config.base_settings import BaseSettings
from zenml.config.global_config import GlobalConfiguration
from zenml.constants import (
    ENV_ZENML_LOCAL_STORES_PATH,
)
from zenml.deployers.base_deployer import (
    BaseDeployerConfig,
    BaseDeployerFlavor,
    BaseDeployerSettings,
)
from zenml.deployers.containerized_deployer import (
    ContainerizedDeployer,
)
from zenml.deployers.exceptions import (
    DeployerError,
    DeploymentDeprovisionError,
    DeploymentLogsNotFoundError,
    DeploymentNotFoundError,
    DeploymentProvisionError,
)
from zenml.deployers.server.entrypoint_configuration import (
    DEPLOYMENT_ID_OPTION,
    DeploymentEntrypointConfiguration,
)
from zenml.enums import DeploymentStatus, StackComponentType
from zenml.logger import get_logger
from zenml.models import (
    DeploymentOperationalState,
    DeploymentResponse,
)
from zenml.stack import Stack, StackValidator
from zenml.utils import docker_utils
from zenml.utils.networking_utils import (
    lookup_preferred_or_free_port,
)

logger = get_logger(__name__)


class DockerDeploymentMetadata(BaseModel):
    """Metadata for a Docker deployment."""

    port: Optional[int] = None
    container_id: Optional[str] = None
    container_name: Optional[str] = None
    container_image_id: Optional[str] = None
    container_image_uri: Optional[str] = None
    container_status: Optional[str] = None

    @classmethod
    def from_container(
        cls, container: Container
    ) -> "DockerDeploymentMetadata":
        """Create a DockerDeploymentMetadata from a docker container.

        Args:
            container: The docker container to get the metadata for.

        Returns:
            The metadata for the docker container.
        """
        image = container.image
        if image is not None:
            image_url = image.attrs["RepoTags"][0]
            image_id = image.attrs["Id"]
        else:
            image_url = None
            image_id = None
        if container.ports:
            ports = list(container.ports.values())
            if len(ports) > 0:
                port = int(ports[0][0]["HostPort"])
            else:
                port = None
        else:
            port = None
        return cls(
            port=port,
            container_id=container.id,
            container_name=container.name,
            container_image_uri=image_url,
            container_image_id=image_id,
            container_status=container.status,
        )

    @classmethod
    def from_deployment(
        cls, deployment: DeploymentResponse
    ) -> "DockerDeploymentMetadata":
        """Create a DockerDeploymentMetadata from a deployment.

        Args:
            deployment: The deployment to get the metadata for.

        Returns:
            The metadata for the deployment.
        """
        return cls.model_validate(deployment.deployment_metadata)


class DockerDeployer(ContainerizedDeployer):
    """Deployer responsible for deploying pipelines locally using Docker."""

    _docker_client: Optional[DockerClient] = None

    @property
    def settings_class(self) -> Optional[Type["BaseSettings"]]:
        """Settings class for the Docker deployer.

        Returns:
            The settings class.
        """
        return DockerDeployerSettings

    @property
    def config(self) -> "DockerDeployerConfig":
        """Returns the `DockerDeployerConfig` config.

        Returns:
            The configuration.
        """
        return cast(DockerDeployerConfig, self._config)

    @property
    def validator(self) -> Optional[StackValidator]:
        """Ensures there is an image builder in the stack.

        Returns:
            A `StackValidator` instance.
        """
        return StackValidator(
            required_components={StackComponentType.IMAGE_BUILDER}
        )

    @property
    def docker_client(self) -> DockerClient:
        """Initialize and/or return the docker client.

        Returns:
            The docker client.
        """
        if self._docker_client is None:
            self._docker_client = (
                docker_utils._try_get_docker_client_from_env()
            )
        return self._docker_client

    def _get_container_id(self, deployment: DeploymentResponse) -> str:
        """Get the docker container id associated with a deployment.

        Args:
            deployment: The deployment to get the container id for.

        Returns:
            The docker container id for the deployment.
        """
        return f"zenml-deployment-{deployment.id}"

    def _get_container(
        self, deployment: DeploymentResponse
    ) -> Optional[Container]:
        """Get the docker container associated with a deployment.

        Args:
            deployment: The deployment to get the container for.

        Returns:
            The docker container for the service, or None if the container
            does not exist.
        """
        try:
            return self.docker_client.containers.get(
                self._get_container_id(deployment)
            )
        except docker_errors.NotFound:
            # container doesn't exist yet or was removed
            return None

    def _get_container_operational_state(
        self, container: Container
    ) -> DeploymentOperationalState:
        """Get the operational state of a docker container running a deployment.

        Args:
            container: The docker container to get the operational state of.

        Returns:
            The operational state of the docker container running the pipeline
            deployment.
        """
        metadata = DockerDeploymentMetadata.from_container(container)
        state = DeploymentOperationalState(
            status=DeploymentStatus.UNKNOWN,
            metadata=metadata.model_dump(exclude_none=True),
        )
        if metadata.container_status == "running":
            state.status = DeploymentStatus.RUNNING
        elif metadata.container_status == "exited":
            state.status = DeploymentStatus.ERROR
        elif metadata.container_status in ["created", "restarting", "paused"]:
            state.status = DeploymentStatus.PENDING
        elif metadata.container_status == "dead":
            state.status = DeploymentStatus.ERROR
        elif metadata.container_status == "removing":
            state.status = DeploymentStatus.PENDING
        elif metadata.container_status == "exited":
            state.status = DeploymentStatus.ABSENT
        elif metadata.container_status == "dead":
            state.status = DeploymentStatus.ERROR

        if state.status == DeploymentStatus.RUNNING:
            state.url = "http://localhost"
            if metadata.port:
                state.url += f":{metadata.port}"

        return state

    def do_provision_deployment(
        self,
        deployment: DeploymentResponse,
        stack: "Stack",
        environment: Dict[str, str],
        secrets: Dict[str, str],
        timeout: int,
    ) -> DeploymentOperationalState:
        """Deploy a pipeline as a Docker container.

        Args:
            deployment: The deployment to run as a Docker container.
            stack: The stack the pipeline will be deployed on.
            environment: A dictionary of environment variables to set on the
                deployment.
            secrets: A dictionary of secret environment variables to set
                on the deployment. These secret environment variables
                should not be exposed as regular environment variables on the
                deployer.
            timeout: The maximum time in seconds to wait for the pipeline
                deployment to be provisioned.

        Returns:
            The DeploymentOperationalState object representing the
            operational state of the provisioned deployment.

        Raises:
            DeploymentProvisionError: if provisioning the deployment
                fails.
        """
        assert deployment.snapshot, "Pipeline snapshot not found"
        snapshot = deployment.snapshot

        # Currently, there is no safe way to pass secrets to a docker
        # container, so we simply merge them into the environment variables.
        environment.update(secrets)

        settings = cast(
            DockerDeployerSettings,
            self.get_settings(snapshot),
        )

        existing_metadata = DockerDeploymentMetadata.from_deployment(
            deployment
        )

        entrypoint = DeploymentEntrypointConfiguration.get_entrypoint_command()

        entrypoint_kwargs = {
            DEPLOYMENT_ID_OPTION: deployment.id,
        }

        arguments = DeploymentEntrypointConfiguration.get_entrypoint_arguments(
            **entrypoint_kwargs
        )

        # Add the local stores path as a volume mount
        stack.check_local_paths()
        local_stores_path = GlobalConfiguration().local_stores_path
        volumes = {
            local_stores_path: {
                "bind": local_stores_path,
                "mode": "rw",
            }
        }
        environment[ENV_ZENML_LOCAL_STORES_PATH] = local_stores_path

        # check if a container already exists for the deployment
        container = self._get_container(deployment)

        if container:
            # the container exists, check if it is running
            if container.status == "running":
                logger.debug(
                    f"Container for deployment '{deployment.name}' is "
                    "already running",
                )
                container.stop(timeout=timeout)

            # the container is stopped or in an error state, remove it
            logger.debug(
                f"Removing previous container for deployment "
                f"'{deployment.name}'",
            )
            container.remove(force=True)

        logger.debug(
            f"Starting container for deployment '{deployment.name}'..."
        )

        image = self.get_image(deployment.snapshot)

        try:
            self.docker_client.images.get(image)
        except docker_errors.ImageNotFound:
            logger.debug(
                f"Pulling container image '{image}' for deployment "
                f"'{deployment.name}'...",
            )
            self.docker_client.images.pull(image)

        preferred_ports: List[int] = []
        if settings.port:
            preferred_ports.append(settings.port)
        if existing_metadata.port:
            preferred_ports.append(existing_metadata.port)
        port = lookup_preferred_or_free_port(
            preferred_ports=preferred_ports,
            allocate_port_if_busy=settings.allocate_port_if_busy,
            range=settings.port_range,
            address="0.0.0.0",  # nosec
        )
        container_port = (
            snapshot.pipeline_configuration.deployment_settings.uvicorn_port
        )
        ports: Dict[str, Optional[int]] = {f"{container_port}/tcp": port}

        uid_args: Dict[str, Any] = {}
        if sys.platform == "win32":
            # File permissions are not checked on Windows. This if clause
            # prevents mypy from complaining about unused 'type: ignore'
            # statements
            pass
        else:
            # Run the container in the context of the local UID/GID
            # to ensure that the local database can be shared
            # with the container.
            logger.debug(
                "Setting UID and GID to local user/group in container."
            )
            uid_args = dict(
                user=os.getuid(),
                group_add=[os.getgid()],
            )

        run_args = copy.deepcopy(settings.run_args)
        docker_environment = run_args.pop("environment", {})
        docker_environment.update(environment)

        docker_volumes = run_args.pop("volumes", {})
        docker_volumes.update(volumes)

        extra_hosts = run_args.pop("extra_hosts", {})
        extra_hosts["host.docker.internal"] = "host-gateway"

        run_args.update(uid_args)

        try:
            container = self.docker_client.containers.run(
                image=image,
                name=self._get_container_id(deployment),
                entrypoint=entrypoint,
                command=arguments,
                detach=True,
                volumes=docker_volumes,
                environment=docker_environment,
                remove=False,
                auto_remove=False,
                ports=ports,
                labels={
                    "zenml-deployment-id": str(deployment.id),
                    "zenml-deployment-name": deployment.name,
                    "zenml-deployer-name": str(self.name),
                    "zenml-deployer-id": str(self.id),
                    "managed-by": "zenml",
                },
                extra_hosts=extra_hosts,
                **run_args,
            )

            logger.debug(
                f"Docker container for deployment '{deployment.name}' "
                f"started with ID {self._get_container_id(deployment)}",
            )

        except docker_errors.DockerException as e:
            raise DeploymentProvisionError(
                f"Docker container for deployment '{deployment.name}' "
                f"failed to start: {e}"
            )

        return self._get_container_operational_state(container)

    def do_get_deployment_state(
        self,
        deployment: DeploymentResponse,
    ) -> DeploymentOperationalState:
        """Get information about a docker deployment.

        Args:
            deployment: The deployment to get information about.

        Returns:
            The DeploymentOperationalState object representing the
            updated operational state of the deployment.

        Raises:
            DeploymentNotFoundError: if no deployment is found
                corresponding to the provided DeploymentResponse.
        """
        container = self._get_container(deployment)
        if container is None:
            raise DeploymentNotFoundError(
                f"Docker container for deployment '{deployment.name}' "
                "not found"
            )

        return self._get_container_operational_state(container)

    def do_get_deployment_state_logs(
        self,
        deployment: DeploymentResponse,
        follow: bool = False,
        tail: Optional[int] = None,
    ) -> Generator[str, bool, None]:
        """Get the logs of a Docker deployment.

        Args:
            deployment: The deployment to get the logs of.
            follow: if True, the logs will be streamed as they are written
            tail: only retrieve the last NUM lines of log output.

        Yields:
            The logs of the deployment.

        Raises:
            DeploymentNotFoundError: if no deployment is found
                corresponding to the provided DeploymentResponse.
            DeploymentLogsNotFoundError: if the deployment logs are not
                found.
            DeployerError: if the deployment logs cannot
                be retrieved for any other reason or if an unexpected error
                occurs.
        """
        container = self._get_container(deployment)
        if container is None:
            raise DeploymentNotFoundError(
                f"Docker container for deployment '{deployment.name}' "
                "not found"
            )

        try:
            log_kwargs: Dict[str, Any] = {
                "stdout": True,
                "stderr": True,
                "stream": follow,
                "follow": follow,
                "timestamps": True,
            }

            if tail is not None and tail > 0:
                log_kwargs["tail"] = tail

            log_stream = container.logs(**log_kwargs)

            if follow:
                for log_line in log_stream:
                    if isinstance(log_line, bytes):
                        yield log_line.decode(
                            "utf-8", errors="replace"
                        ).rstrip()
                    else:
                        yield str(log_line).rstrip()
            else:
                if isinstance(log_stream, bytes):
                    log_text = log_stream.decode("utf-8", errors="replace")
                    for line in log_text.splitlines():
                        yield line
                else:
                    for log_line in log_stream:
                        if isinstance(log_line, bytes):
                            yield log_line.decode(
                                "utf-8", errors="replace"
                            ).rstrip()
                        else:
                            yield str(log_line).rstrip()

        except docker_errors.NotFound as e:
            raise DeploymentLogsNotFoundError(
                f"Logs for deployment '{deployment.name}' not found: {e}"
            )
        except docker_errors.APIError as e:
            raise DeployerError(
                f"Docker API error while retrieving logs for deployment "
                f"'{deployment.name}': {e}"
            )
        except docker_errors.DockerException as e:
            raise DeployerError(
                f"Docker error while retrieving logs for deployment "
                f"'{deployment.name}': {e}"
            )
        except Exception as e:
            raise DeployerError(
                f"Unexpected error while retrieving logs for deployment "
                f"'{deployment.name}': {e}"
            )

    def do_deprovision_deployment(
        self,
        deployment: DeploymentResponse,
        timeout: int,
    ) -> Optional[DeploymentOperationalState]:
        """Deprovision a docker deployment.

        Args:
            deployment: The deployment to deprovision.
            timeout: The maximum time in seconds to wait for the pipeline
                deployment to be deprovisioned.

        Returns:
            The DeploymentOperationalState object representing the
            operational state of the deleted deployment, or None if the
            deletion is completed before the call returns.

        Raises:
            DeploymentNotFoundError: if no deployment is found
                corresponding to the provided DeploymentResponse.
            DeploymentDeprovisionError: if the deployment
                deprovision fails.
        """
        container = self._get_container(deployment)
        if container is None:
            raise DeploymentNotFoundError(
                f"Docker container for deployment '{deployment.name}' "
                "not found"
            )

        try:
            container.stop(timeout=timeout)
            container.remove()
        except docker_errors.DockerException as e:
            raise DeploymentDeprovisionError(
                f"Docker container for deployment '{deployment.name}' "
                f"failed to delete: {e}"
            )

        return None


class DockerDeployerSettings(BaseDeployerSettings):
    """Docker deployer settings.

    Attributes:
        port: The port to expose the deployment on.
        allocate_port_if_busy: If True, allocate a free port if the configured
            port is busy.
        port_range: The range of ports to search for a free port.
        run_args: Arguments to pass to the `docker run` call. (See
            https://docker-py.readthedocs.io/en/stable/containers.html for a list
            of what can be passed.)
    """

    port: Optional[int] = None
    allocate_port_if_busy: bool = True
    port_range: Tuple[int, int] = (8000, 65535)
    run_args: Dict[str, Any] = {}


class DockerDeployerConfig(BaseDeployerConfig, DockerDeployerSettings):
    """Docker deployer config."""

    @property
    def is_local(self) -> bool:
        """Checks if this stack component is running locally.

        Returns:
            True if this config is for a local component, False otherwise.
        """
        return True


class DockerDeployerFlavor(BaseDeployerFlavor):
    """Flavor for the Docker deployer."""

    @property
    def name(self) -> str:
        """Name of the deployer flavor.

        Returns:
            Name of the deployer flavor.
        """
        return "docker"

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/deployer/docker.png"

    @property
    def config_class(self) -> Type[BaseDeployerConfig]:
        """Config class for the base deployer flavor.

        Returns:
            The config class.
        """
        return DockerDeployerConfig

    @property
    def implementation_class(self) -> Type["DockerDeployer"]:
        """Implementation class for this flavor.

        Returns:
            Implementation class for this flavor.
        """
        return DockerDeployer



# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

# Minor parts of the `prepare_or_run_pipeline()` method of this file are
# inspired by the kubeflow dag runner implementation of tfx
"""Implementation of the Kubeflow orchestrator."""

import os
import types
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    cast,
)
from uuid import UUID

import kfp
import requests
import urllib3
from kfp import dsl
from kfp.client import Client as KFPClient
from kfp.compiler import Compiler
from kfp_server_api.exceptions import ApiException
from kubernetes import client as k8s_client
from kubernetes import config as k8s_config

from zenml.client import Client
from zenml.config.resource_settings import ResourceSettings
from zenml.constants import (
    METADATA_ORCHESTRATOR_URL,
)
from zenml.entrypoints import StepEntrypointConfiguration
from zenml.enums import StackComponentType
from zenml.environment import Environment
from zenml.integrations.kubeflow.flavors.kubeflow_orchestrator_flavor import (
    KubeflowOrchestratorConfig,
    KubeflowOrchestratorSettings,
)
from zenml.io import fileio
from zenml.logger import get_logger
from zenml.metadata.metadata_types import MetadataType, Uri
from zenml.orchestrators import ContainerizedOrchestrator, SubmissionResult
from zenml.orchestrators.utils import get_orchestrator_run_name
from zenml.stack import StackValidator
from zenml.utils import io_utils, settings_utils

if TYPE_CHECKING:
    from zenml.models import PipelineRunResponse, PipelineSnapshotResponse
    from zenml.stack import Stack


logger = get_logger(__name__)

KFP_POD_LABELS = {
    "add-pod-env": "true",
    "pipelines.kubeflow.org/pipeline-sdk-type": "zenml",
}

ENV_KFP_RUN_ID = "KFP_RUN_ID"
KFP_ACCELERATOR_NODE_SELECTOR_CONSTRAINT_LABEL = "accelerator"


class KubeClientKFPClient(kfp.Client):  # type: ignore[misc]
    """KFP client initialized from a Kubernetes client.

    This is a workaround for the fact that the native KFP client does not
    support initialization from an existing Kubernetes client.
    """

    def __init__(
        self, client: k8s_client.ApiClient, *args: Any, **kwargs: Any
    ) -> None:
        """Initializes the KFP client from a Kubernetes client.

        Args:
            client: pre-configured Kubernetes client.
            args: standard KFP client positional arguments.
            kwargs: standard KFP client keyword arguments.
        """
        self._k8s_client = client
        super().__init__(*args, **kwargs)

    def _load_config(self, *args: Any, **kwargs: Any) -> Any:
        """Loads the KFP configuration.

        Initializes the KFP configuration from the Kubernetes client.

        Args:
            args: standard KFP client positional arguments.
            kwargs: standard KFP client keyword arguments.

        Returns:
            The KFP configuration.
        """
        from kfp_server_api.configuration import Configuration

        kube_config = self._k8s_client.configuration

        host = (
            kube_config.host
            + "/"
            + self._KUBE_PROXY_PATH.format(kwargs.get("namespace", "kubeflow"))
        )

        config = Configuration(
            host=host,
            api_key=kube_config.api_key,
            api_key_prefix=kube_config.api_key_prefix,
            username=kube_config.username,
            password=kube_config.password,
            discard_unknown_keys=kube_config.discard_unknown_keys,
        )

        # Extra attributes not present in the Configuration constructor
        keys = ["ssl_ca_cert", "cert_file", "key_file", "verify_ssl"]
        for key in keys:
            if key in kube_config.__dict__:
                setattr(config, key, getattr(kube_config, key))

        return config


class KubeflowOrchestrator(ContainerizedOrchestrator):
    """Orchestrator responsible for running pipelines using Kubeflow."""

    _k8s_client: Optional[k8s_client.ApiClient] = None

    def _get_kfp_client(
        self,
        settings: KubeflowOrchestratorSettings,
    ) -> kfp.Client:
        """Creates a KFP client instance.

        Args:
            settings: Settings which can be used to
                configure the client instance.

        Returns:
            A KFP client instance.

        Raises:
            RuntimeError: If the linked Kubernetes connector behaves
                unexpectedly.
        """
        connector = self.get_connector()
        client_args = settings.client_args.copy()

        # The kube_context, host and namespace are stack component
        # configurations that refer to the Kubeflow deployment. We don't want
        # these overwritten on a run by run basis by user settings
        client_args["namespace"] = self.config.kubeflow_namespace

        if connector:
            client = connector.connect()
            if not isinstance(client, k8s_client.ApiClient):
                raise RuntimeError(
                    f"Expected a k8s_client.ApiClient while trying to use the "
                    f"linked connector, but got {type(client)}."
                )
            return KubeClientKFPClient(
                client=client,
                **client_args,
            )

        elif self.config.kubernetes_context:
            client_args["kube_context"] = self.config.kubernetes_context

        elif self.config.kubeflow_hostname:
            client_args["host"] = self.config.kubeflow_hostname

            # Handle username and password, ignore the case if one is passed and
            # not the other. Also do not attempt to get cookie if cookie is
            # already passed in client_args
            if settings.client_username and settings.client_password:
                # If cookie is already set, then ignore
                if "cookie" in client_args:
                    logger.warning(
                        "Cookie already set in `client_args`, ignoring "
                        "`client_username` and `client_password`..."
                    )
                else:
                    session_cookie = self._get_session_cookie(
                        username=settings.client_username,
                        password=settings.client_password,
                    )

                    client_args["cookies"] = session_cookie
        return KFPClient(**client_args)

    @property
    def config(self) -> KubeflowOrchestratorConfig:
        """Returns the `KubeflowOrchestratorConfig` config.

        Returns:
            The configuration.
        """
        return cast(KubeflowOrchestratorConfig, self._config)

    def get_kubernetes_contexts(self) -> Tuple[List[str], Optional[str]]:
        """Get the list of configured Kubernetes contexts and the active context.

        Returns:
            A tuple containing the list of configured Kubernetes contexts and
            the active context.
        """
        try:
            contexts, active_context = k8s_config.list_kube_config_contexts()
        except k8s_config.config_exception.ConfigException:
            return [], None

        context_names = [c["name"] for c in contexts]
        active_context_name = active_context["name"]
        return context_names, active_context_name

    @property
    def settings_class(self) -> Type[KubeflowOrchestratorSettings]:
        """Settings class for the Kubeflow orchestrator.

        Returns:
            The settings class.
        """
        return KubeflowOrchestratorSettings

    @property
    def validator(self) -> Optional[StackValidator]:
        """Validates that the stack contains a container registry.

        Also check that requirements are met for local components.

        Returns:
            A `StackValidator` instance.
        """
        msg = f"'{self.name}' Kubeflow orchestrator error: "

        def _validate_kube_context(
            kubernetes_context: str,
        ) -> Tuple[bool, str]:
            contexts, active_context = self.get_kubernetes_contexts()

            if kubernetes_context and kubernetes_context not in contexts:
                if not self.config.is_local:
                    return False, (
                        f"{msg}could not find a Kubernetes context named "
                        f"'{kubernetes_context}' in the local Kubernetes "
                        f"configuration. Please make sure that the Kubernetes "
                        f"cluster is running and that the kubeconfig file is "
                        f"configured correctly. To list all configured "
                        f"contexts, run:\n\n"
                        f"  `kubectl config get-contexts`\n"
                    )
            elif (
                kubernetes_context
                and active_context
                and kubernetes_context != active_context
            ):
                logger.warning(
                    f"{msg}the Kubernetes context '{kubernetes_context}' "  # nosec
                    f"configured for the Kubeflow orchestrator is not the "
                    f"same as the active context in the local Kubernetes "
                    f"configuration. If this is not deliberate, you should "
                    f"update the orchestrator's `kubernetes_context` field by "
                    f"running:\n\n"
                    f"  `zenml orchestrator update {self.name} "
                    f"--kubernetes_context={active_context}`\n"
                    f"To list all configured contexts, run:\n\n"
                    f"  `kubectl config get-contexts`\n"
                    f"To set the active context to be the same as the one "
                    f"configured in the Kubeflow orchestrator and silence "
                    f"this warning, run:\n\n"
                    f"  `kubectl config use-context "
                    f"{kubernetes_context}`\n"
                )

            return True, ""

        def _validate_local_requirements(stack: "Stack") -> Tuple[bool, str]:
            container_registry = stack.container_registry

            # should not happen, because the stack validation takes care of
            # this, but just in case
            assert container_registry is not None

            kubernetes_context = self.config.kubernetes_context

            if not self.connector:
                if (
                    not kubernetes_context
                    and not self.config.kubeflow_hostname
                ):
                    return False, (
                        f"{msg}the Kubeflow orchestrator is incompletely "
                        "configured. For a multi-tenant Kubeflow deployment, "
                        "you must set the `kubeflow_hostname` attribute in the "
                        "orchestrator configuration. For a single-tenant "
                        "deployment, you must either set the "
                        "`kubernetes_context` attribute in the orchestrator "
                        "configuration to the name of the Kubernetes config "
                        "context pointing to the cluster where you would like "
                        "to run pipelines or link this stack component to a "
                        "Kubernetes cluster via a service connector (see the "
                        "'zenml orchestrator connect' CLI command)."
                    )

                if kubernetes_context:
                    valid, err = _validate_kube_context(kubernetes_context)
                    if not valid:
                        return False, err

            silence_local_validations_msg = (
                f"To silence this warning, set the "
                f"`skip_local_validations` attribute to True in the "
                f"orchestrator configuration by running:\n\n"
                f"  'zenml orchestrator update {self.name} "
                f"--skip_local_validations=True'\n"
            )

            if not self.config.is_local:
                # if the orchestrator is not running in a local k3d cluster,
                # we cannot have any other local components in our stack,
                # because we cannot mount the local path into the container.
                # This may result in problems when running the pipeline,
                # because the local components will not be available inside the
                # Kubeflow containers.

                # go through all stack components and identify those that
                # advertise a local path where they persist information that
                # they need to be available when running pipelines.
                for stack_comp in stack.components.values():
                    local_path = stack_comp.local_path
                    if not local_path:
                        continue
                    return False, (
                        f"{msg}the Kubeflow orchestrator is configured to run "
                        f"pipelines in a remote Kubernetes cluster but the "
                        f"'{stack_comp.name}' {stack_comp.type.value} is a "
                        "local stack component and will not be available in "
                        "the Kubeflow pipeline step.\n"
                        "Please ensure that you always use non-local "
                        f"stack components with a remote Kubeflow "
                        f"orchestrator, otherwise you may run into pipeline "
                        f"execution problems. You should use a flavor of "
                        f"{stack_comp.type.value} other than "
                        f"'{stack_comp.flavor}'.\n"
                        + silence_local_validations_msg
                    )

                # if the orchestrator is remote, the container registry must
                # also be remote.
                if container_registry.config.is_local:
                    return False, (
                        f"{msg}the Kubeflow orchestrator is configured to run "
                        f"pipelines in a remote Kubernetes cluster, but the "
                        f"'{container_registry.name}' container registry URI "
                        f"'{container_registry.config.uri}' "
                        f"points to a local container registry. Please ensure "
                        f"that you always use non-local stack components with "
                        f"a remote Kubeflow orchestrator, otherwise you will "
                        f"run into problems. You should use a flavor of "
                        f"container registry other than "
                        f"'{container_registry.flavor}'.\n"
                        + silence_local_validations_msg
                    )

            return True, ""

        return StackValidator(
            required_components={
                StackComponentType.CONTAINER_REGISTRY,
                StackComponentType.IMAGE_BUILDER,
            },
            custom_validation_function=_validate_local_requirements,
        )

    @property
    def root_directory(self) -> str:
        """Path to the root directory for all files concerning this orchestrator.

        Returns:
            Path to the root directory.
        """
        return os.path.join(
            io_utils.get_global_config_directory(),
            "kubeflow",
            str(self.id),
        )

    @property
    def pipeline_directory(self) -> str:
        """Returns path to a directory in which the kubeflow pipeline files are stored.

        Returns:
            Path to the pipeline directory.
        """
        return os.path.join(self.root_directory, "pipelines")

    def _create_dynamic_component(
        self,
        image: str,
        command: List[str],
        arguments: List[str],
        component_name: str,
    ) -> dsl.PipelineTask:
        """Creates a dynamic container component for a Kubeflow pipeline.

        Args:
            image: The image to use for the component.
            command: The command to use for the component.
            arguments: The arguments to use for the component.
            component_name: The name of the component.

        Returns:
            The dynamic container component.
        """

        def dynamic_container_component() -> dsl.ContainerSpec:
            """Dynamic container component.

            Returns:
                The dynamic container component.
            """
            return dsl.ContainerSpec(
                image=image,
                command=command,
                args=arguments,
            )

        # Change the name of the function
        new_container_spec_func = types.FunctionType(
            dynamic_container_component.__code__,
            dynamic_container_component.__globals__,
            name=component_name,
            argdefs=dynamic_container_component.__defaults__,
            closure=dynamic_container_component.__closure__,
        )
        pipeline_task = dsl.container_component(new_container_spec_func)

        return pipeline_task

    def submit_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        stack: "Stack",
        base_environment: Dict[str, str],
        step_environments: Dict[str, Dict[str, str]],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submits a pipeline to the orchestrator.

        This method should only submit the pipeline and not wait for it to
        complete. If the orchestrator is configured to wait for the pipeline run
        to complete, a function that waits for the pipeline run to complete can
        be passed as part of the submission result.

        Args:
            snapshot: The pipeline snapshot to submit.
            stack: The stack the pipeline will run on.
            base_environment: Base environment shared by all steps. This should
                be set if your orchestrator for example runs one container that
                is responsible for starting all the steps.
            step_environments: Environment variables to set when executing
                specific steps.
            placeholder_run: An optional placeholder run for the snapshot.

        Raises:
            RuntimeError: If trying to run a pipeline in a notebook
                environment.

        Returns:
            Optional submission result.
        """
        # First check whether the code running in a notebook
        if Environment.in_notebook():
            raise RuntimeError(
                "The Kubeflow orchestrator cannot run pipelines in a notebook "
                "environment. The reason is that it is non-trivial to create "
                "a Docker image of a notebook. Please consider refactoring "
                "your notebook cells into separate scripts in a Python module "
                "and run the code outside of a notebook when using this "
                "orchestrator."
            )

        assert stack.container_registry

        # Create a callable for future compilation into a dsl.Pipeline.
        orchestrator_run_name = get_orchestrator_run_name(
            pipeline_name=snapshot.pipeline_configuration.name
        ).replace("_", "-")

        def _create_dynamic_pipeline() -> Any:
            """Create a dynamic pipeline including each step.

            Returns:
                pipeline_func
            """
            step_name_to_dynamic_component: Dict[str, Any] = {}
            node_selector_constraint: Optional[Tuple[str, str]] = None

            for step_name, step in snapshot.step_configurations.items():
                image = self.get_image(
                    snapshot=snapshot,
                    step_name=step_name,
                )
                command = StepEntrypointConfiguration.get_entrypoint_command()
                arguments = (
                    StepEntrypointConfiguration.get_entrypoint_arguments(
                        step_name=step_name,
                        snapshot_id=snapshot.id,
                    )
                )
                dynamic_component = self._create_dynamic_component(
                    image, command, arguments, step_name
                )
                step_settings = cast(
                    KubeflowOrchestratorSettings, self.get_settings(step)
                )
                pod_settings = step_settings.pod_settings
                if pod_settings:
                    ignored_fields = pod_settings.model_fields_set - {
                        "node_selectors"
                    }
                    if ignored_fields:
                        logger.warning(
                            f"The following pod settings are not supported in "
                            f"Kubeflow with Kubeflow Pipelines 2.x and will be "
                            f"ignored: {list(ignored_fields)}."
                        )

                    # apply pod settings
                    if (
                        KFP_ACCELERATOR_NODE_SELECTOR_CONSTRAINT_LABEL
                        in pod_settings.node_selectors.keys()
                    ):
                        node_selector_constraint = (
                            KFP_ACCELERATOR_NODE_SELECTOR_CONSTRAINT_LABEL,
                            pod_settings.node_selectors[
                                KFP_ACCELERATOR_NODE_SELECTOR_CONSTRAINT_LABEL
                            ],
                        )

                step_name_to_dynamic_component[step_name] = dynamic_component

            @dsl.pipeline(  # type: ignore[misc]
                display_name=orchestrator_run_name,
            )
            def dynamic_pipeline() -> None:
                """Dynamic pipeline."""
                # iterate through the components one by one
                # (from step_name_to_dynamic_component)
                for (
                    component_name,
                    component,
                ) in step_name_to_dynamic_component.items():
                    step_environment = step_environments[component_name]
                    # for each component, check to see what other steps are
                    # upstream of it
                    step = snapshot.step_configurations[component_name]
                    upstream_step_components = [
                        step_name_to_dynamic_component[upstream_step_name]
                        for upstream_step_name in step.spec.upstream_steps
                    ]
                    task = (
                        component()
                        .set_display_name(
                            name=component_name,
                        )
                        .set_caching_options(enable_caching=False)
                        .set_env_variable(
                            name=ENV_KFP_RUN_ID,
                            value=dsl.PIPELINE_JOB_NAME_PLACEHOLDER,
                        )
                        .after(*upstream_step_components)
                    )
                    for key, value in step_environment.items():
                        task = task.set_env_variable(name=key, value=value)

                    self._configure_container_resources(
                        task,
                        step.config.resource_settings,
                        node_selector_constraint,
                    )

            return dynamic_pipeline

        # Get a filepath to use to save the finished yaml to
        fileio.makedirs(self.pipeline_directory)
        pipeline_file_path = os.path.join(
            self.pipeline_directory, f"{orchestrator_run_name}.yaml"
        )

        # write the argo pipeline yaml
        Compiler().compile(
            pipeline_func=_create_dynamic_pipeline(),
            package_path=pipeline_file_path,
            pipeline_name=orchestrator_run_name,
        )

        logger.info(
            "Writing Kubeflow workflow definition to `%s`.", pipeline_file_path
        )

        # using the kfp client uploads the pipeline to kubeflow pipelines and
        # runs it there
        return self._upload_and_run_pipeline(
            snapshot=snapshot,
            pipeline_file_path=pipeline_file_path,
            run_name=orchestrator_run_name,
        )

    def _upload_and_run_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        pipeline_file_path: str,
        run_name: str,
    ) -> Optional[SubmissionResult]:
        """Tries to upload and run a KFP pipeline.

        Args:
            snapshot: The pipeline snapshot.
            pipeline_file_path: Path to the pipeline definition file.
            run_name: The Kubeflow run name.

        Raises:
            RuntimeError: If Kubeflow API returns an error.

        Returns:
            Optional submission result.
        """
        pipeline_name = snapshot.pipeline_configuration.name
        settings = cast(
            KubeflowOrchestratorSettings, self.get_settings(snapshot)
        )
        user_namespace = settings.user_namespace

        kubernetes_context = self.config.kubernetes_context
        try:
            if kubernetes_context:
                logger.info(
                    "Running in kubernetes context '%s'.",
                    kubernetes_context,
                )
            elif self.config.kubeflow_hostname:
                logger.info(
                    "Running on Kubeflow deployment '%s'.",
                    self.config.kubeflow_hostname,
                )
            elif self.connector:
                logger.info(
                    "Running with Kubernetes credentials from connector '%s'.",
                    str(self.connector),
                )

            # upload the pipeline to Kubeflow and start it

            client = self._get_kfp_client(settings=settings)
            if snapshot.schedule:
                try:
                    experiment = client.get_experiment(
                        pipeline_name, namespace=user_namespace
                    )
                    logger.info(
                        "A recurring run has already been created with this "
                        "pipeline. Creating new recurring run now.."
                    )
                except (ValueError, ApiException):
                    experiment = client.create_experiment(
                        pipeline_name, namespace=user_namespace
                    )
                    logger.info(
                        "Creating a new recurring run for pipeline '%s'.. ",
                        pipeline_name,
                    )
                logger.info(
                    "You can see all recurring runs under the '%s' experiment.",
                    pipeline_name,
                )

                interval_seconds = (
                    snapshot.schedule.interval_second.seconds
                    if snapshot.schedule.interval_second
                    else None
                )
                result = client.create_recurring_run(
                    experiment_id=experiment.experiment_id,
                    job_name=run_name,
                    pipeline_package_path=pipeline_file_path,
                    enable_caching=False,
                    cron_expression=snapshot.schedule.cron_expression,
                    start_time=snapshot.schedule.utc_start_time,
                    end_time=snapshot.schedule.utc_end_time,
                    interval_second=interval_seconds,
                    no_catchup=not snapshot.schedule.catchup,
                )

                logger.info(
                    "Started recurring run with ID '%s'.",
                    result.recurring_run_id,
                )
            else:
                logger.info(
                    "No schedule detected. Creating a one-off pipeline run.."
                )
                try:
                    result = client.create_run_from_pipeline_package(
                        pipeline_file_path,
                        arguments={},
                        run_name=run_name,
                        enable_caching=False,
                        namespace=user_namespace,
                    )
                except ApiException:
                    raise RuntimeError(
                        f"Failed to create {run_name} on kubeflow! "
                        "Please check stack component settings and "
                        "configuration!"
                    )

                logger.info(
                    "Started one-off pipeline run with ID '%s'.", result.run_id
                )

                if settings.synchronous:

                    def _wait_for_completion() -> None:
                        client.wait_for_run_completion(
                            run_id=result.run_id, timeout=settings.timeout
                        )

                    return SubmissionResult(
                        wait_for_completion=_wait_for_completion
                    )
        except urllib3.exceptions.HTTPError as error:
            if kubernetes_context:
                msg = (
                    f"Please make sure your kubernetes config is present and "
                    f"the '{kubernetes_context}' kubernetes context is "
                    "configured correctly."
                )
            elif self.connector:
                msg = (
                    f"Please check that the '{self.connector}' connector "
                    f"linked to this component is configured correctly with "
                    "valid credentials."
                )
            else:
                msg = ""

            logger.warning(
                f"Failed to upload Kubeflow pipeline: {error}. {msg}",
            )

        return None

    def get_orchestrator_run_id(self) -> str:
        """Returns the active orchestrator run id.

        Raises:
            RuntimeError: If the environment variable specifying the run id
                is not set.

        Returns:
            The orchestrator run id.
        """
        try:
            return os.environ[ENV_KFP_RUN_ID]
        except KeyError:
            raise RuntimeError(
                "Unable to read run id from environment variable "
                f"{ENV_KFP_RUN_ID}."
            )

    def _get_session_cookie(self, username: str, password: str) -> str:
        """Gets session cookie from username and password.

        Args:
            username: Username for kubeflow host.
            password: Password for kubeflow host.

        Raises:
            RuntimeError: If the cookie fetching failed.

        Returns:
            Cookie with the prefix `authsession=`.
        """
        if self.config.kubeflow_hostname is None:
            raise RuntimeError(
                "You must configure the Kubeflow orchestrator "
                "with the `kubeflow_hostname` parameter which usually ends "
                "with `/pipeline` (e.g. `https://mykubeflow.com/pipeline`). "
                "Please update the current kubeflow orchestrator with: "
                f"`zenml orchestrator update {self.name} "
                "--kubeflow_hostname=<MY_KUBEFLOW_HOST>`"
            )

        # Get cookie
        logger.info(
            f"Attempting to fetch session cookie from {self.config.kubeflow_hostname} "
            "with supplied username and password..."
        )
        session = requests.Session()
        try:
            response = session.get(self.config.kubeflow_hostname)
            response.raise_for_status()
        except (
            requests.exceptions.HTTPError,
            requests.exceptions.ConnectionError,
            requests.exceptions.Timeout,
            requests.exceptions.RequestException,
        ) as e:
            raise RuntimeError(
                f"Error while trying to fetch kubeflow cookie: {e}"
            )

        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
        }
        data = {"login": username, "password": password}
        try:
            response = session.post(response.url, headers=headers, data=data)
            response.raise_for_status()
        except requests.exceptions.HTTPError as errh:
            raise RuntimeError(
                f"Error while trying to fetch kubeflow cookie: {errh}"
            )

        cookie_dict: Dict[str, str] = session.cookies.get_dict()  # type: ignore[no-untyped-call, unused-ignore]

        if "authservice_session" not in cookie_dict:
            raise RuntimeError("Invalid username and/or password!")

        logger.info("Session cookie fetched successfully!")

        return "authservice_session=" + str(cookie_dict["authservice_session"])

    def get_pipeline_run_metadata(
        self, run_id: UUID
    ) -> Dict[str, "MetadataType"]:
        """Get general component-specific metadata for a pipeline run.

        Args:
            run_id: The ID of the pipeline run.

        Returns:
            A dictionary of metadata.
        """
        hostname = self.config.kubeflow_hostname
        if not hostname:
            return {}

        hostname = hostname.rstrip("/")
        pipeline_suffix = "/pipeline"
        if hostname.endswith(pipeline_suffix):
            hostname = hostname[: -len(pipeline_suffix)]

        run = Client().get_pipeline_run(run_id)

        settings_key = settings_utils.get_stack_component_setting_key(self)
        run_settings = self.settings_class.model_validate(
            run.config.model_dump().get(settings_key, self.config)
        )
        user_namespace = run_settings.user_namespace

        if user_namespace:
            run_url = (
                f"{hostname}/_/pipeline/?ns={user_namespace}#"
                f"/runs/details/{self.get_orchestrator_run_id()}"
            )
            return {
                METADATA_ORCHESTRATOR_URL: Uri(run_url),
            }
        else:
            return {
                METADATA_ORCHESTRATOR_URL: Uri(f"{hostname}"),
            }

    def _configure_container_resources(
        self,
        dynamic_component: dsl.PipelineTask,
        resource_settings: "ResourceSettings",
        node_selector_constraint: Optional[Tuple[str, str]] = None,
    ) -> dsl.PipelineTask:
        """Adds resource requirements to the container.

        Args:
            dynamic_component: The dynamic component to add the resource
                settings to.
            resource_settings: The resource settings to use for this
                container.
            node_selector_constraint: Node selector constraint to apply to
                the container.

        Returns:
            The dynamic component with the resource settings applied.
        """
        # Set optional CPU, RAM and GPU constraints for the pipeline
        if resource_settings:
            cpu_limit = resource_settings.cpu_count or None

        if cpu_limit is not None:
            dynamic_component = dynamic_component.set_cpu_limit(str(cpu_limit))

        memory_limit = resource_settings.get_memory() or None
        if memory_limit is not None:
            dynamic_component = dynamic_component.set_memory_limit(
                memory_limit
            )

        gpu_limit = (
            resource_settings.gpu_count
            if resource_settings.gpu_count is not None
            else 0
        )

        if node_selector_constraint:
            (constraint_label, value) = node_selector_constraint
            if gpu_limit is not None and gpu_limit > 0:
                dynamic_component = (
                    dynamic_component.set_accelerator_type(value)
                    .set_accelerator_limit(gpu_limit)
                    .set_gpu_limit(gpu_limit)
                )
            elif constraint_label == "accelerator" and gpu_limit == 0:
                logger.warning(
                    "GPU limit is set to 0 but a GPU type is specified. Ignoring GPU settings."
                )

        return dynamic_component



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Run:AI step operator implementation."""

import random
import time
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, cast

from runai.models.annotation import Annotation
from runai.models.environment_variable import EnvironmentVariable
from runai.models.extended_resource import ExtendedResource
from runai.models.image_pull_secret import ImagePullSecret
from runai.models.label import Label
from runai.models.superset_spec_all_of_compute import SupersetSpecAllOfCompute
from runai.models.toleration import Toleration
from runai.models.training_creation_request import TrainingCreationRequest
from runai.models.training_spec_spec import TrainingSpecSpec

from zenml.config.base_settings import BaseSettings
from zenml.config.build_configuration import BuildConfiguration
from zenml.enums import ExecutionStatus, StackComponentType
from zenml.integrations.runai.client.runai_client import (
    RunAIClient,
    RunAIClientError,
    RunAIWorkloadNotFoundError,
)
from zenml.integrations.runai.constants import (
    MAX_WORKLOAD_NAME_LENGTH,
    is_failure_status,
    is_pending_status,
    is_success_status,
    map_runai_status_to_execution_status,
)
from zenml.integrations.runai.flavors.runai_step_operator_flavor import (
    RunAIStepOperatorConfig,
    RunAIStepOperatorSettings,
)
from zenml.logger import get_logger
from zenml.orchestrators.publish_utils import publish_step_run_metadata
from zenml.stack import Stack, StackValidator
from zenml.step_operators import BaseStepOperator

if TYPE_CHECKING:
    from zenml.config.step_run_info import StepRunInfo
    from zenml.models import PipelineSnapshotBase, StepRunResponse

logger = get_logger(__name__)

RUNAI_STEP_OPERATOR_DOCKER_IMAGE_KEY = "runai_step_operator"
RUNAI_WORKLOAD_ID_METADATA_KEY = "workload_id"
RUNAI_WORKLOAD_NAME_METADATA_KEY = "workload_name"


class RunAIStepOperator(BaseStepOperator):
    """Step operator to run individual steps on Run:AI.

    This step operator enables selective GPU offloading by running
    individual pipeline steps on Run:AI clusters.

    Example usage:
    ```python
    @step(step_operator="runai")
    def train_model(data):
        # GPU-intensive training runs on Run:AI
        ...

    ```
    """

    _client: Optional[RunAIClient] = None

    @property
    def client(self) -> RunAIClient:
        """Get or create the Run:AI client.

        The client is cached for reuse across multiple calls.

        Returns:
            The RunAIClient instance.
        """
        if self._client is None:
            self._client = RunAIClient(
                client_id=self.config.client_id.get_secret_value(),
                client_secret=self.config.client_secret.get_secret_value(),
                runai_base_url=self.config.runai_base_url,
            )
        return self._client

    @property
    def config(self) -> RunAIStepOperatorConfig:
        """Returns the step operator config.

        Returns:
            The configuration.
        """
        return cast(RunAIStepOperatorConfig, self._config)

    @property
    def settings_class(self) -> Optional[Type[BaseSettings]]:
        """Settings class for the Run:AI step operator.

        Returns:
            The settings class.
        """
        return RunAIStepOperatorSettings

    @property
    def validator(self) -> Optional[StackValidator]:
        """Validates the stack.

        Returns:
            A validator that checks that the stack contains a remote container
            registry and a remote artifact store.
        """

        def _validate_remote_components(stack: Stack) -> Tuple[bool, str]:
            if stack.artifact_store.config.is_local:
                return False, (
                    "The Run:AI step operator runs code remotely and "
                    "needs to write files into the artifact store, but the "
                    f"artifact store `{stack.artifact_store.name}` of the "
                    "active stack is local. Please ensure that your stack "
                    "contains a remote artifact store when using the Run:AI "
                    "step operator."
                )

            container_registry = stack.container_registry
            assert container_registry is not None

            if container_registry.config.is_local:
                return False, (
                    "The Run:AI step operator runs code remotely and "
                    "needs to push/pull Docker images, but the "
                    f"container registry `{container_registry.name}` of the "
                    "active stack is local. Please ensure that your stack "
                    "contains a remote container registry when using the "
                    "Run:AI step operator."
                )

            return True, ""

        return StackValidator(
            required_components={
                StackComponentType.CONTAINER_REGISTRY,
                StackComponentType.IMAGE_BUILDER,
            },
            custom_validation_function=_validate_remote_components,
        )

    def get_docker_builds(
        self, snapshot: "PipelineSnapshotBase"
    ) -> List[BuildConfiguration]:
        """Gets the Docker builds required for the component.

        Args:
            snapshot: The pipeline snapshot for which to get the builds.

        Returns:
            The required Docker builds.
        """
        builds = []
        for step_name, step in snapshot.step_configurations.items():
            if step.config.uses_step_operator(self.name):
                build = BuildConfiguration(
                    key=RUNAI_STEP_OPERATOR_DOCKER_IMAGE_KEY,
                    settings=step.config.docker_settings,
                    step_name=step_name,
                )
                builds.append(build)

        return builds

    def submit(
        self,
        info: "StepRunInfo",
        entrypoint_command: List[str],
        environment: Dict[str, str],
    ) -> None:
        """Submits a step to Run:AI as a training workload.

        Args:
            info: Information about the step run.
            entrypoint_command: Command that executes the step.
            environment: Environment variables to set in the step operator
                environment.

        Raises:
            RunAIClientError: If building the Run:AI training request fails.
            RuntimeError: If workload submission fails.
        """
        settings = cast(RunAIStepOperatorSettings, self.get_settings(info))

        image = info.get_image(key=RUNAI_STEP_OPERATOR_DOCKER_IMAGE_KEY)

        project_id, cluster_id = self._resolve_project_and_cluster()

        workload_name = self._build_workload_name(info)

        compute = self._build_compute_spec(settings)

        env_vars = self._build_environment_variables(environment)

        image_pull_secrets = self._build_image_pull_secrets()

        command, args = self._build_command_and_args(entrypoint_command)

        tolerations_list = self._build_tolerations(settings)
        labels_list = self._build_labels(settings)
        annotations_list = self._build_annotations(settings)

        try:
            training_request = TrainingCreationRequest(
                name=workload_name,
                project_id=project_id,
                cluster_id=cluster_id,
                spec=TrainingSpecSpec(
                    image=image,
                    command=command,
                    compute=compute,
                    environment_variables=env_vars,
                    args=args,
                    image_pull_secrets=image_pull_secrets,
                    node_pools=settings.node_pools,
                    node_type=settings.node_type,
                    preemptibility=settings.preemptibility,
                    priority_class=settings.priority_class,
                    tolerations=tolerations_list,
                    backoff_limit=settings.backoff_limit,
                    termination_grace_period_seconds=(
                        settings.termination_grace_period_seconds
                    ),
                    terminate_after_preemption=settings.terminate_after_preemption,
                    working_dir=settings.working_dir,
                    labels=labels_list,
                    annotations=annotations_list,
                ),
            )
        except Exception as exc:
            raise RunAIClientError(
                "Failed to build Run:AI training request "
                f"({type(exc).__name__}): {exc}"
            ) from exc

        info.force_write_logs()

        try:
            result = self.client.create_training_workload(training_request)
            logger.info(
                "Submitted step '%s' to Run:AI as workload '%s' (ID: %s)",
                info.pipeline_step_name,
                result.workload_name,
                result.workload_id,
            )
        except RunAIClientError as exc:
            raise RuntimeError(
                f"Failed to submit step '{info.pipeline_step_name}' to Run:AI: {exc}. "
                "Verify credentials, project name, cluster access, and quota."
            ) from exc

        publish_step_run_metadata(
            info.step_run_id,
            {
                self.id: {
                    RUNAI_WORKLOAD_ID_METADATA_KEY: result.workload_id,
                    RUNAI_WORKLOAD_NAME_METADATA_KEY: result.workload_name,
                }
            },
        )
        info.step_run.run_metadata[RUNAI_WORKLOAD_ID_METADATA_KEY] = (
            result.workload_id
        )
        info.step_run.run_metadata[RUNAI_WORKLOAD_NAME_METADATA_KEY] = (
            result.workload_name
        )

    def _get_workload_id(self, step_run: "StepRunResponse") -> str:
        """Gets the Run:AI workload ID from step run metadata.

        Args:
            step_run: The step run.

        Returns:
            The workload ID.

        Raises:
            RuntimeError: If the workload metadata is missing.
        """
        workload_id = step_run.run_metadata.get(RUNAI_WORKLOAD_ID_METADATA_KEY)
        if not workload_id:
            raise RuntimeError(
                "Unable to determine the Run:AI workload ID for step run "
                f"`{step_run.id}` because run metadata key "
                f"`{RUNAI_WORKLOAD_ID_METADATA_KEY}` is missing."
            )
        return str(workload_id)

    def get_status(self, step_run: "StepRunResponse") -> ExecutionStatus:
        """Gets the status of a submitted step.

        Args:
            step_run: The step run.

        Returns:
            The step status.
        """
        workload_id = self._get_workload_id(step_run)
        try:
            status = self.client.get_training_workload_status(workload_id)
        except RunAIWorkloadNotFoundError:
            logger.warning(
                "Run:AI workload `%s` for step run `%s` was not found.",
                workload_id,
                step_run.id,
            )
            return ExecutionStatus.FAILED

        if status is None:
            logger.warning(
                "Run:AI workload `%s` for step run `%s` has no status.",
                workload_id,
                step_run.id,
            )
            return ExecutionStatus.FAILED

        return map_runai_status_to_execution_status(status)

    def cancel(self, step_run: "StepRunResponse") -> None:
        """Cancels a submitted step.

        Args:
            step_run: The step run.
        """
        workload_id = self._get_workload_id(step_run)
        self.client.suspend_training_workload(workload_id)

    def wait(self, step_run: "StepRunResponse") -> ExecutionStatus:
        """Waits for a submitted step to finish.

        Args:
            step_run: The step run.

        Returns:
            The final step status.
        """
        settings = cast(RunAIStepOperatorSettings, self.get_settings(step_run))
        workload_id = self._get_workload_id(step_run)
        status = self._wait_for_completion(
            client=self.client,
            workload_id=workload_id,
            settings=settings,
        )
        logger.info("Run:AI step operator job completed.")
        return status

    def _resolve_project_and_cluster(self) -> Tuple[str, str]:
        """Resolve Run:AI project and cluster IDs.

        Returns:
            Tuple of (project_id, cluster_id).

        Raises:
            RuntimeError: If project or cluster cannot be resolved.
        """
        try:
            project = self.client.get_project_by_name(self.config.project_name)
        except RunAIClientError as exc:
            raise RuntimeError(str(exc)) from exc

        cluster_id = project.cluster_id

        if self.config.cluster_name and cluster_id:
            cluster = self.client.get_cluster_by_id(cluster_id)
            if cluster.name != self.config.cluster_name:
                logger.warning(
                    f"Configured cluster '{self.config.cluster_name}' "
                    f"does not match project's cluster '{cluster.name}'. "
                    f"Using project's cluster."
                )

        if not cluster_id:
            try:
                if self.config.cluster_name:
                    cluster = self.client.get_cluster_by_name(
                        self.config.cluster_name
                    )
                    cluster_id = cluster.id
                else:
                    cluster = self.client.get_first_cluster()
                    cluster_id = cluster.id
            except RunAIClientError as exc:
                raise RuntimeError(str(exc)) from exc

        return project.id, cluster_id

    def _sanitize_name_component(self, name: str) -> str:
        """Sanitize a name component for Kubernetes DNS label compliance.

        Args:
            name: The name component to sanitize.

        Returns:
            A sanitized string with only lowercase alphanumeric chars and hyphens.
        """
        import re

        sanitized = name.lower()
        sanitized = re.sub(r"[^a-z0-9-]", "-", sanitized)
        sanitized = re.sub(r"-+", "-", sanitized)
        return sanitized.strip("-")

    def _build_workload_name(self, info: "StepRunInfo") -> str:
        """Build a unique workload name for the step.

        Run:AI workload names must be valid Kubernetes DNS labels:
        lowercase, start with letter, alphanumeric and hyphens only.

        Args:
            info: Step run information.

        Returns:
            str: A unique workload name that conforms to RunAI naming
                requirements.
        """
        step_name = self._sanitize_name_component(info.pipeline_step_name)
        pipeline_name = self._sanitize_name_component(info.pipeline.name)

        base_name = f"zenml-{pipeline_name}-{step_name}"

        if base_name and not base_name[0].isalpha():
            base_name = f"z{base_name}"

        run_id_suffix = str(info.run_id)[:8]
        timestamp = datetime.now(timezone.utc).strftime("%H%M%S%f")[:9]

        identifier_len = len(run_id_suffix) + len(timestamp) + 2
        max_base_len = MAX_WORKLOAD_NAME_LENGTH - identifier_len
        base_name = base_name[:max_base_len].rstrip("-")

        return f"{base_name}-{run_id_suffix}-{timestamp}"

    def _build_command_and_args(
        self, entrypoint_command: List[str]
    ) -> Tuple[str, Optional[str]]:
        """Build the command and arguments for Run:AI.

        Run:AI expects the command and args as strings (not lists). We keep
        the interpreter/module invocation intact for the command and join
        remaining tokens as the args string.

        Args:
            entrypoint_command: The full entrypoint command list.

        Returns:
            Tuple of (command, args) as strings. The args value is None when
            there are no extra tokens.

        Raises:
            ValueError: If entrypoint_command format is invalid.
        """
        if len(entrypoint_command) < 3:
            raise ValueError(
                f"Expected entrypoint command with at least 3 elements "
                f"(e.g., ['python', '-m', 'module_name']), but got "
                f"{len(entrypoint_command)} elements: {entrypoint_command}"
            )

        command_tokens = entrypoint_command[:3]
        args_tokens = entrypoint_command[3:]

        command = " ".join(command_tokens)
        args = " ".join(args_tokens) if args_tokens else None
        return command, args

    def _build_environment_variables(
        self, environment: Dict[str, str]
    ) -> Optional[List[EnvironmentVariable]]:
        """Build environment variables for the Run:AI workload.

        Args:
            environment: Mapping of environment variable names to values.

        Returns:
            List of EnvironmentVariable objects or None.
        """
        if not environment:
            return None

        return [
            EnvironmentVariable(name=name, value=str(value))
            for name, value in environment.items()
        ]

    def _build_compute_spec(
        self, settings: RunAIStepOperatorSettings
    ) -> SupersetSpecAllOfCompute:
        """Build the compute specification for the workload.

        Args:
            settings: The step operator settings.

        Returns:
            SupersetSpecAllOfCompute: The compute specification object used for
                RunAI workload submission.
        """
        extended_resources_list = None
        if settings.extended_resources:
            extended_resources_list = [
                ExtendedResource(resource=k, quantity=v)
                for k, v in settings.extended_resources.items()
            ]

        return SupersetSpecAllOfCompute(
            gpu_devices_request=settings.gpu_devices_request,
            gpu_portion_request=settings.gpu_portion_request,
            gpu_request_type=settings.gpu_request_type,
            gpu_memory_request=settings.gpu_memory_request,
            gpu_portion_limit=settings.gpu_portion_limit,
            gpu_memory_limit=settings.gpu_memory_limit,
            cpu_core_request=settings.cpu_core_request,
            cpu_core_limit=settings.cpu_core_limit,
            cpu_memory_request=settings.cpu_memory_request,
            cpu_memory_limit=settings.cpu_memory_limit,
            large_shm_request=settings.large_shm_request,
            extended_resources=extended_resources_list,
        )

    def _build_image_pull_secrets(
        self,
    ) -> Optional[List[ImagePullSecret]]:
        """Build image pull secrets for the workload.

        Returns:
            List of ImagePullSecret objects or None.
        """
        if not self.config.image_pull_secret_name:
            return None

        logger.info(
            f"Using image pull secret '{self.config.image_pull_secret_name}'. "
            f"Ensure this secret exists in your Run:AI project."
        )

        return [
            ImagePullSecret(
                name=self.config.image_pull_secret_name, user_credential=False
            )
        ]

    def _build_tolerations(
        self, settings: RunAIStepOperatorSettings
    ) -> Optional[List[Toleration]]:
        """Build tolerations for scheduling on tainted nodes.

        Args:
            settings: The step operator settings.

        Returns:
            List of Toleration objects or None.
        """
        if not settings.tolerations:
            return None

        tolerations_list = []
        for t in settings.tolerations:
            toleration_dict: Dict[str, Any] = {}
            if "key" in t:
                toleration_dict["key"] = t["key"]
            if "operator" in t:
                toleration_dict["operator"] = t["operator"]
            if "value" in t:
                toleration_dict["value"] = t["value"]
            if "effect" in t:
                toleration_dict["effect"] = t["effect"]
            tolerations_list.append(Toleration(**toleration_dict))

        return tolerations_list

    def _build_labels(
        self, settings: RunAIStepOperatorSettings
    ) -> Optional[List[Label]]:
        """Build labels for the workload pod.

        Args:
            settings: The step operator settings.

        Returns:
            List of Label objects or None.
        """
        if not settings.labels:
            return None

        return [Label(name=k, value=v) for k, v in settings.labels.items()]

    def _build_annotations(
        self, settings: RunAIStepOperatorSettings
    ) -> Optional[List[Annotation]]:
        """Build annotations for the workload pod.

        Args:
            settings: The step operator settings.

        Returns:
            List of Annotation objects or None.
        """
        if not settings.annotations:
            return None

        return [
            Annotation(name=k, value=v)
            for k, v in settings.annotations.items()
        ]

    def _cleanup_workload(
        self, client: RunAIClient, workload_id: str, reason: str
    ) -> None:
        """Stop a Run:AI workload and optionally delete it.

        Args:
            client: The RunAIClient instance.
            workload_id: The workload ID to stop.
            reason: Reason for cleanup (for logging).
        """
        manual_cleanup_hint = (
            "Open "
            f"{self.config.runai_base_url} and navigate to Workloads > "
            f"Training to clean up workload '{workload_id}'."
        )
        logger.info(f"Stopping workload {workload_id}: {reason}")
        try:
            client.suspend_training_workload(workload_id)
        except Exception as suspend_exc:
            logger.error(
                f"Failed to stop workload {workload_id}: {suspend_exc}. "
                f"{manual_cleanup_hint}"
            )

        if not self.config.delete_on_failure:
            logger.info(
                f"Preserving workload {workload_id} after failure "
                f"(delete_on_failure=False). Reason: {reason}. "
                f"{manual_cleanup_hint}"
            )
            return

        logger.info(f"Deleting workload {workload_id}: {reason}")
        try:
            client.delete_training_workload(workload_id)
        except Exception as delete_exc:
            logger.error(
                f"Failed to delete workload {workload_id}: {delete_exc}. "
                f"{manual_cleanup_hint}"
            )

    def _wait_for_completion(
        self,
        client: RunAIClient,
        workload_id: str,
        settings: RunAIStepOperatorSettings,
    ) -> ExecutionStatus:
        """Wait for a Run:AI workload to complete.

        Args:
            client: The RunAIClient instance.
            workload_id: The workload ID to wait for.
            settings: The step operator settings.

        Returns:
            The final status of the workload.

        Raises:
            RuntimeError: If the workload fails or times out.
        """
        start_time = time.time()
        timeout = settings.workload_timeout
        pending_timeout = settings.pending_timeout
        retry_count = 0
        max_retries = 3
        base_interval = self.config.monitoring_interval
        missing_status_retries = 0
        max_missing_status_checks = 3
        pending_start_time: Optional[float] = None

        while True:
            sleep_time = base_interval
            if timeout and (time.time() - start_time) > timeout:
                self._cleanup_workload(client, workload_id, "timeout")
                raise RuntimeError(
                    f"Run:AI workload {workload_id} timed out after {timeout} seconds"
                )

            try:
                status = client.get_training_workload_status(workload_id)
                retry_count = 0

                if status is None:
                    missing_status_retries += 1
                    if missing_status_retries > max_missing_status_checks:
                        raise RuntimeError(
                            "Run:AI workload "
                            f"{workload_id} is missing or returned no status "
                            f"after {max_missing_status_checks} checks. The "
                            "workload might have been deleted or failed to "
                            "start."
                        )

                    logger.warning(
                        "No status returned for Run:AI workload %s "
                        "(attempt %d/%d); retrying.",
                        workload_id,
                        missing_status_retries,
                        max_missing_status_checks,
                    )
                    sleep_time = base_interval
                elif status and is_success_status(status):
                    missing_status_retries = 0
                    return ExecutionStatus.COMPLETED

                elif status and is_failure_status(status):
                    missing_status_retries = 0
                    self._cleanup_workload(
                        client, workload_id, f"failed with status: {status}"
                    )
                    raise RuntimeError(
                        f"Run:AI workload {workload_id} failed with status: {status}"
                    )

                else:
                    missing_status_retries = 0
                    if status and is_pending_status(status):
                        if pending_start_time is None:
                            pending_start_time = time.time()
                        if (
                            pending_timeout
                            and (time.time() - pending_start_time)
                            > pending_timeout
                        ):
                            self._cleanup_workload(
                                client,
                                workload_id,
                                f"pending timeout after {pending_timeout} seconds",
                            )
                            raise RuntimeError(
                                "Run:AI workload "
                                f"{workload_id} exceeded pending timeout of "
                                f"{pending_timeout} seconds"
                            )
                    else:
                        pending_start_time = None
                    sleep_time = base_interval
            except RunAIWorkloadNotFoundError as exc:
                retry_count = 0
                missing_status_retries += 1
                if missing_status_retries > max_missing_status_checks:
                    raise RuntimeError(
                        "Run:AI workload "
                        f"{workload_id} was not found after "
                        f"{max_missing_status_checks} checks. The workload "
                        "might have been deleted or failed to start."
                    ) from exc

                logger.warning(
                    "Run:AI workload %s not found (attempt %d/%d); retrying.",
                    workload_id,
                    missing_status_retries,
                    max_missing_status_checks,
                )
                sleep_time = base_interval
            except RunAIClientError as exc:
                retry_count += 1
                if retry_count > max_retries:
                    self._cleanup_workload(
                        client, workload_id, "max retries exceeded"
                    )
                    raise RuntimeError(
                        f"Failed to check status after {max_retries} retries: {exc}"
                    )

                sleep_time = min(base_interval * (2**retry_count), 300)
                sleep_time += random.uniform(0, 0.1 * sleep_time)
                logger.warning(
                    f"Status check failed (retry {retry_count}/{max_retries}): {exc}"
                )

            time.sleep(sleep_time)



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Cache policy."""

from typing import Any, List, Optional, Union

from pydantic import BaseModel, BeforeValidator, Field, field_validator
from typing_extensions import Annotated

from zenml.config.source import Source, SourceWithValidator
from zenml.logger import get_logger

logger = get_logger(__name__)


class CachePolicy(BaseModel):
    """Cache policy."""

    include_step_code: bool = Field(
        default=True,
        description="Whether to include the step code in the cache key.",
    )
    include_step_parameters: bool = Field(
        default=True,
        description="Whether to include the step parameters in the cache key.",
    )
    include_artifact_values: bool = Field(
        default=True,
        description="Whether to include the artifact values in the cache key. "
        "If the materializer for an artifact doesn't support generating a "
        "content hash, the artifact ID will be used as a fallback if enabled.",
    )
    include_artifact_ids: bool = Field(
        default=True,
        description="Whether to include the artifact IDs in the cache key.",
    )
    ignored_inputs: Optional[List[str]] = Field(
        default=None,
        description="List of input names to ignore in the cache key.",
    )
    file_dependencies: Optional[List[str]] = Field(
        default=None,
        description="List of file paths. The contents of theses files will be "
        "included in the cache key. Only relative paths within the source root "
        "are allowed.",
    )
    source_dependencies: Optional[List[SourceWithValidator]] = Field(
        default=None,
        description="List of Python objects (modules, classes, functions). "
        "The source code of these objects will be included in the cache key.",
    )
    cache_func: Optional[SourceWithValidator] = Field(
        default=None,
        description="Function without arguments that returns a string. The "
        "returned value will be included in the cache key.",
    )
    expires_after: Optional[int] = Field(
        default=None,
        description="The number of seconds after which the cached result by a "
        "step with this cache policy will expire. If not set, the result "
        "will never expire.",
    )

    @field_validator("source_dependencies", mode="before")
    def _validate_source_dependencies(
        cls, v: Optional[List[Any]]
    ) -> Optional[List[Any]]:
        from zenml.utils import source_utils

        if v is None:
            return None

        result = []
        for obj in v:
            if isinstance(obj, (str, Source, dict)):
                result.append(obj)
            else:
                result.append(source_utils.resolve(obj))
        return result

    @field_validator("cache_func", mode="before")
    def _validate_cache_func(cls, v: Optional[Any]) -> Optional[Any]:
        from zenml.utils import source_utils

        if v is None or isinstance(v, (str, Source, dict)):
            return v
        elif callable(v):
            return source_utils.resolve(v)
        else:
            raise ValueError(f"Invalid cache function: {v}")

    @classmethod
    def default(cls) -> "CachePolicy":
        """Default cache policy.

        Returns:
            The default cache policy.
        """
        return cls(
            include_step_code=True,
            include_step_parameters=True,
            include_artifact_values=True,
            include_artifact_ids=True,
            ignored_inputs=None,
        )

    @classmethod
    def from_string(cls, value: str) -> "CachePolicy":
        """Create a cache policy from a string.

        Args:
            value: The string to create a cache policy from.

        Raises:
            ValueError: If the string is not a valid cache policy.

        Returns:
            The cache policy.
        """
        if value.lower() == "default":
            return cls.default()
        else:
            raise ValueError(f"Invalid cache policy: {value}")


def _convert_cache_policy(value: Any) -> Any:
    """Converts a potential cache policy string to a cache policy object.

    Args:
        value: The value to convert.

    Returns:
        The converted value.
    """
    if isinstance(value, str):
        return CachePolicy.from_string(value)

    return value


CachePolicyWithValidator = Annotated[
    CachePolicy, BeforeValidator(_convert_cache_policy)
]
CachePolicyOrString = Union[CachePolicy, str]



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Models representing devices."""

from datetime import datetime
from typing import Optional, Union
from uuid import UUID

from pydantic import Field

from zenml.enums import OAuthDeviceStatus
from zenml.models.v2.base.base import (
    BaseRequest,
    BaseUpdate,
)
from zenml.models.v2.base.scoped import (
    UserScopedFilter,
    UserScopedResponse,
    UserScopedResponseBody,
    UserScopedResponseMetadata,
    UserScopedResponseResources,
)

# ------------------ Request Model ------------------


class OAuthDeviceInternalRequest(BaseRequest):
    """Internal request model for OAuth2 devices."""

    client_id: UUID = Field(description="The client ID of the OAuth2 device.")
    expires_in: int = Field(
        description="The number of seconds after which the OAuth2 device "
        "expires and can no longer be used for authentication."
    )
    os: Optional[str] = Field(
        default=None,
        description="The operating system of the device used for "
        "authentication.",
    )
    ip_address: Optional[str] = Field(
        default=None,
        description="The IP address of the device used for authentication.",
    )
    hostname: Optional[str] = Field(
        default=None,
        description="The hostname of the device used for authentication.",
    )
    python_version: Optional[str] = Field(
        default=None,
        description="The Python version of the device used for authentication.",
    )
    zenml_version: Optional[str] = Field(
        default=None,
        description="The ZenML version of the device used for authentication.",
    )
    city: Optional[str] = Field(
        default=None,
        description="The city where the device is located.",
    )
    region: Optional[str] = Field(
        default=None,
        description="The region where the device is located.",
    )
    country: Optional[str] = Field(
        default=None,
        description="The country where the device is located.",
    )


# ------------------ Update Model ------------------


class OAuthDeviceUpdate(BaseUpdate):
    """OAuth2 device update model."""

    locked: Optional[bool] = Field(
        default=None,
        description="Whether to lock or unlock the OAuth2 device. A locked "
        "device cannot be used for authentication.",
    )


class OAuthDeviceInternalUpdate(OAuthDeviceUpdate):
    """OAuth2 device update model used internally for authentication."""

    user_id: Optional[UUID] = Field(
        default=None, description="User that owns the OAuth2 device."
    )
    status: Optional[OAuthDeviceStatus] = Field(
        default=None, description="The new status of the OAuth2 device."
    )
    expires_in: Optional[int] = Field(
        default=None,
        description="Set the device to expire in the given number of seconds. "
        "If the value is 0 or negative, the device is set to never expire.",
    )
    failed_auth_attempts: Optional[int] = Field(
        default=None,
        description="Set the number of failed authentication attempts.",
    )
    trusted_device: Optional[bool] = Field(
        default=None,
        description="Whether to mark the OAuth2 device as trusted. A trusted "
        "device has a much longer validity time.",
    )
    update_last_login: bool = Field(
        default=False, description="Whether to update the last login date."
    )
    generate_new_codes: bool = Field(
        default=False,
        description="Whether to generate new user and device codes.",
    )
    os: Optional[str] = Field(
        default=None,
        description="The operating system of the device used for "
        "authentication.",
    )
    ip_address: Optional[str] = Field(
        default=None,
        description="The IP address of the device used for authentication.",
    )
    hostname: Optional[str] = Field(
        default=None,
        description="The hostname of the device used for authentication.",
    )
    python_version: Optional[str] = Field(
        default=None,
        description="The Python version of the device used for authentication.",
    )
    zenml_version: Optional[str] = Field(
        default=None,
        description="The ZenML version of the device used for authentication.",
    )
    city: Optional[str] = Field(
        default=None,
        description="The city where the device is located.",
    )
    region: Optional[str] = Field(
        default=None,
        description="The region where the device is located.",
    )
    country: Optional[str] = Field(
        default=None,
        description="The country where the device is located.",
    )


# ------------------ Response Model ------------------


class OAuthDeviceResponseBody(UserScopedResponseBody):
    """Response body for OAuth2 devices."""

    client_id: UUID = Field(description="The client ID of the OAuth2 device.")
    expires: Optional[datetime] = Field(
        default=None,
        description="The expiration date of the OAuth2 device after which "
        "the device is no longer valid and cannot be used for "
        "authentication.",
    )
    trusted_device: bool = Field(
        description="Whether the OAuth2 device was marked as trusted. A "
        "trusted device has a much longer validity time.",
    )
    status: OAuthDeviceStatus = Field(
        description="The status of the OAuth2 device."
    )
    os: Optional[str] = Field(
        default=None,
        description="The operating system of the device used for "
        "authentication.",
    )
    ip_address: Optional[str] = Field(
        default=None,
        description="The IP address of the device used for authentication.",
    )
    hostname: Optional[str] = Field(
        default=None,
        description="The hostname of the device used for authentication.",
    )


class OAuthDeviceResponseMetadata(UserScopedResponseMetadata):
    """Response metadata for OAuth2 devices."""

    python_version: Optional[str] = Field(
        default=None,
        description="The Python version of the device used for authentication.",
    )
    zenml_version: Optional[str] = Field(
        default=None,
        description="The ZenML version of the device used for authentication.",
    )
    city: Optional[str] = Field(
        default=None,
        description="The city where the device is located.",
    )
    region: Optional[str] = Field(
        default=None,
        description="The region where the device is located.",
    )
    country: Optional[str] = Field(
        default=None,
        description="The country where the device is located.",
    )
    failed_auth_attempts: int = Field(
        description="The number of failed authentication attempts.",
    )
    last_login: Optional[datetime] = Field(
        description="The date of the last successful login."
    )


class OAuthDeviceResponseResources(UserScopedResponseResources):
    """Class for all resource models associated with the OAuthDevice entity."""


class OAuthDeviceResponse(
    UserScopedResponse[
        OAuthDeviceResponseBody,
        OAuthDeviceResponseMetadata,
        OAuthDeviceResponseResources,
    ]
):
    """Response model for OAuth2 devices."""

    _warn_on_response_updates = False

    def get_hydrated_version(self) -> "OAuthDeviceResponse":
        """Get the hydrated version of this OAuth2 device.

        Returns:
            an instance of the same entity with the metadata field attached.
        """
        from zenml.client import Client

        return Client().zen_store.get_authorized_device(self.id)

    # Body and metadata properties
    @property
    def client_id(self) -> UUID:
        """The `client_id` property.

        Returns:
            the value of the property.
        """
        return self.get_body().client_id

    @property
    def expires(self) -> Optional[datetime]:
        """The `expires` property.

        Returns:
            the value of the property.
        """
        return self.get_body().expires

    @property
    def trusted_device(self) -> bool:
        """The `trusted_device` property.

        Returns:
            the value of the property.
        """
        return self.get_body().trusted_device

    @property
    def status(self) -> OAuthDeviceStatus:
        """The `status` property.

        Returns:
            the value of the property.
        """
        return self.get_body().status

    @property
    def os(self) -> Optional[str]:
        """The `os` property.

        Returns:
            the value of the property.
        """
        return self.get_body().os

    @property
    def ip_address(self) -> Optional[str]:
        """The `ip_address` property.

        Returns:
            the value of the property.
        """
        return self.get_body().ip_address

    @property
    def hostname(self) -> Optional[str]:
        """The `hostname` property.

        Returns:
            the value of the property.
        """
        return self.get_body().hostname

    @property
    def python_version(self) -> Optional[str]:
        """The `python_version` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().python_version

    @property
    def zenml_version(self) -> Optional[str]:
        """The `zenml_version` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().zenml_version

    @property
    def city(self) -> Optional[str]:
        """The `city` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().city

    @property
    def region(self) -> Optional[str]:
        """The `region` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().region

    @property
    def country(self) -> Optional[str]:
        """The `country` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().country

    @property
    def failed_auth_attempts(self) -> int:
        """The `failed_auth_attempts` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().failed_auth_attempts

    @property
    def last_login(self) -> Optional[datetime]:
        """The `last_login` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().last_login


class OAuthDeviceInternalResponse(OAuthDeviceResponse):
    """OAuth2 device response model used internally for authentication."""

    user_code: str = Field(
        title="The user code.",
    )
    device_code: str = Field(
        title="The device code.",
    )

    def _verify_code(
        self,
        code: str,
        code_hash: Optional[str],
    ) -> bool:
        """Verifies a given code against the stored (hashed) code.

        Args:
            code: The code to verify.
            code_hash: The hashed code to verify against.

        Returns:
            True if the code is valid, False otherwise.
        """
        from passlib.context import CryptContext

        context = CryptContext(schemes=["bcrypt"], deprecated="auto")
        result = context.verify(code, code_hash)

        return result

    def verify_user_code(
        self,
        user_code: str,
    ) -> bool:
        """Verifies a given user code against the stored (hashed) user code.

        Args:
            user_code: The user code to verify.

        Returns:
            True if the user code is valid, False otherwise.
        """
        return self._verify_code(user_code, self.user_code)

    def verify_device_code(
        self,
        device_code: str,
    ) -> bool:
        """Verifies a given device code against the stored (hashed) device code.

        Args:
            device_code: The device code to verify.

        Returns:
            True if the device code is valid, False otherwise.
        """
        return self._verify_code(device_code, self.device_code)


# ------------------ Filter Model ------------------


class OAuthDeviceFilter(UserScopedFilter):
    """Model to enable advanced filtering of OAuth2 devices."""

    expires: Optional[Union[datetime, str, None]] = Field(
        default=None,
        description="The expiration date of the OAuth2 device.",
        union_mode="left_to_right",
    )
    client_id: Union[UUID, str, None] = Field(
        default=None,
        description="The client ID of the OAuth2 device.",
        union_mode="left_to_right",
    )
    status: Union[OAuthDeviceStatus, str, None] = Field(
        default=None,
        description="The status of the OAuth2 device.",
        union_mode="left_to_right",
    )
    trusted_device: Union[bool, str, None] = Field(
        default=None,
        description="Whether the OAuth2 device was marked as trusted.",
        union_mode="left_to_right",
    )
    failed_auth_attempts: Union[int, str, None] = Field(
        default=None,
        description="The number of failed authentication attempts.",
        union_mode="left_to_right",
    )
    last_login: Optional[Union[datetime, str, None]] = Field(
        default=None,
        description="The date of the last successful login.",
        union_mode="left_to_right",
    )



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the vLLM Inference Server Service."""

import argparse
import os
from typing import Any, List, Optional, Union

from zenml.constants import DEFAULT_LOCAL_SERVICE_IP_ADDRESS
from zenml.logger import get_logger
from zenml.models.v2.misc.service import ServiceType
from zenml.services import (
    HTTPEndpointHealthMonitor,
    HTTPEndpointHealthMonitorConfig,
    LocalDaemonService,
    LocalDaemonServiceConfig,
    LocalDaemonServiceEndpoint,
    LocalDaemonServiceEndpointConfig,
    ServiceEndpointProtocol,
)
from zenml.services.service import BaseDeploymentService

logger = get_logger(__name__)


VLLM_PREDICTION_URL_PATH = "v1"
VLLM_HEALTHCHECK_URL_PATH = "health"


class VLLMDeploymentEndpointConfig(LocalDaemonServiceEndpointConfig):
    """vLLM deployment service configuration.

    Attributes:
        prediction_url_path: URI subpath for prediction requests
    """

    prediction_url_path: str


class VLLMDeploymentEndpoint(LocalDaemonServiceEndpoint):
    """A service endpoint exposed by the vLLM deployment daemon.

    Attributes:
        config: service endpoint configuration
    """

    config: VLLMDeploymentEndpointConfig
    monitor: HTTPEndpointHealthMonitor

    @property
    def prediction_url(self) -> Optional[str]:
        """Gets the prediction URL for the endpoint.

        Returns:
            the prediction URL for the endpoint
        """
        uri = self.status.uri
        if not uri:
            return None
        return os.path.join(uri, self.config.prediction_url_path)


class VLLMServiceConfig(LocalDaemonServiceConfig):
    """vLLM service configurations."""

    model: str
    port: int
    host: Optional[str] = None
    blocking: bool = True
    # If unspecified, model name or path will be used.
    tokenizer: Optional[str] = None
    served_model_name: Optional[Union[str, List[str]]] = None
    # Trust remote code from huggingface.
    trust_remote_code: Optional[bool] = False
    # ['auto', 'slow', 'mistral']
    tokenizer_mode: Optional[str] = "auto"
    # ['auto', 'half', 'float16', 'bfloat16', 'float', 'float32']
    dtype: Optional[str] = "auto"
    # The specific model version to use. It can be a branch name, a tag name, or a commit id.
    # If unspecified, will use the default version.
    revision: Optional[str] = None


class VLLMDeploymentService(LocalDaemonService, BaseDeploymentService):
    """vLLM Inference Server Deployment Service."""

    SERVICE_TYPE = ServiceType(
        name="vllm-deployment",
        type="model-serving",
        flavor="vllm",
        description="vLLM Inference prediction service",
    )
    config: VLLMServiceConfig
    endpoint: VLLMDeploymentEndpoint

    def __init__(self, config: VLLMServiceConfig, **attrs: Any):
        """Initialize the vLLM deployment service.

        Args:
            config: service configuration
            attrs: additional attributes to set on the service
        """
        if isinstance(config, VLLMServiceConfig) and "endpoint" not in attrs:
            endpoint = VLLMDeploymentEndpoint(
                config=VLLMDeploymentEndpointConfig(
                    protocol=ServiceEndpointProtocol.HTTP,
                    port=config.port,
                    ip_address=config.host or DEFAULT_LOCAL_SERVICE_IP_ADDRESS,
                    prediction_url_path=VLLM_PREDICTION_URL_PATH,
                ),
                monitor=HTTPEndpointHealthMonitor(
                    config=HTTPEndpointHealthMonitorConfig(
                        healthcheck_uri_path=VLLM_HEALTHCHECK_URL_PATH,
                    )
                ),
            )
            attrs["endpoint"] = endpoint
        super().__init__(config=config, **attrs)

    def run(self) -> None:
        """Start the service."""
        logger.info(
            "Starting vLLM inference server service as blocking "
            "process... press CTRL+C once to stop it."
        )

        self.endpoint.prepare_for_start()

        import uvloop
        from vllm.entrypoints.openai.api_server import (
            run_server,
        )
        from vllm.entrypoints.openai.cli_args import (
            make_arg_parser,
        )
        from vllm.utils import (
            FlexibleArgumentParser,
        )

        try:
            parser: argparse.ArgumentParser = make_arg_parser(
                FlexibleArgumentParser()
            )
            # pass in empty list to get default args
            # otherwise it will try to get the args from sys.argv
            # and if there's a --config in there, it will want to use
            # that file for vLLM configuration, which is not what we want
            args: argparse.Namespace = parser.parse_args(args=[])
            # Override port with the available port
            self.config.port = self.endpoint.status.port or self.config.port

            # Update the arguments in place
            args.__dict__.update(self.config.model_dump())
            uvloop.run(run_server(args=args))
        except KeyboardInterrupt:
            logger.info("Stopping vLLM prediction service...")

    @property
    def prediction_url(self) -> Optional[str]:
        """Gets the prediction URL for the endpoint.

        Returns:
            the prediction URL for the endpoint
        """
        if not self.is_running:
            return None
        return self.endpoint.prediction_url

    def predict(self, data: "Any") -> "Any":
        """Make a prediction using the service.

        Args:
            data: data to make a prediction on

        Returns:
            The prediction result.

        Raises:
            Exception: if the service is not running
            ValueError: if the prediction endpoint is unknown.
        """
        if not self.is_running:
            raise Exception(
                "vLLM Inference service is not running. "
                "Please start the service before making predictions."
            )
        if self.endpoint.prediction_url is not None:
            from openai import OpenAI

            client = OpenAI(
                api_key="EMPTY",
                base_url=self.endpoint.prediction_url,
            )
            models = client.models.list()
            model = models.data[0].id
            result = client.completions.create(model=model, prompt=data)
            # TODO: We can add support for client.chat.completions.create
        else:
            raise ValueError("No endpoint known for prediction.")
        return result



#  Copyright (c) ZenML GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Logger implementation."""

import logging
import os
import re
import sys
from contextvars import ContextVar
from typing import Any, Dict, Optional

from rich.traceback import install as rich_tb_install

from zenml.constants import (
    ENABLE_RICH_TRACEBACK,
    ENV_ZENML_LOGGING_COLORS_DISABLED,
    ENV_ZENML_SUPPRESS_LOGS,
    ZENML_LOGGING_VERBOSITY,
    ZENML_STORAGE_LOGGING_VERBOSITY,
    handle_bool_env_var,
)
from zenml.enums import LoggingLevels

ZENML_LOGGING_COLORS_DISABLED = handle_bool_env_var(
    ENV_ZENML_LOGGING_COLORS_DISABLED, False
)

step_names_in_console: ContextVar[bool] = ContextVar(
    "step_names_in_console", default=False
)

_original_stdout_write: Optional[Any] = None
_original_stderr_write: Optional[Any] = None
_stdout_wrapped: bool = False
_stderr_wrapped: bool = False


class _ZenMLStdoutStream:
    """Stream that writes to the original stdout, bypassing the ZenML wrapper.

    This ensures console logging doesn't trigger the LoggingContext wrapper,
    preventing duplicate log entries in stored logs.
    """

    def write(self, text: str) -> Any:
        """Write text to the original stdout.

        Args:
            text: The text to write.

        Returns:
            The number of characters written.
        """
        if _original_stdout_write:
            return _original_stdout_write(text)
        return sys.stdout.write(text)

    def flush(self) -> None:
        """Flush the stdout buffer."""
        sys.stdout.flush()


def get_logger(logger_name: str) -> logging.Logger:
    """Main function to get logger name,.

    Args:
        logger_name: Name of logger to initialize.

    Returns:
        A logger object.
    """
    return logging.getLogger(logger_name)


def _add_step_name_to_message(message: str) -> str:
    """Adds the step name to the message.

    Args:
        message: The message to add the step name to.

    Returns:
        The message with the step name added.
    """
    try:
        if step_names_in_console.get():
            from zenml.steps import get_step_context

            step_context = get_step_context()

            if step_context and message not in ["\n", ""]:
                # For progress bar updates (with \r), inject the step name after the \r
                if "\r" in message:
                    message = message.replace(
                        "\r", f"\r[{step_context.step_name}] "
                    )
                else:
                    message = f"[{step_context.step_name}] {message}"
    except Exception:
        # If we can't get step context, just use the original message
        pass

    return message


class ConsoleFormatter(logging.Formatter):
    """Formats logs according to custom specifications."""

    grey: str = "\x1b[90m"
    white: str = "\x1b[37m"
    pink: str = "\x1b[35m"
    green: str = "\x1b[32m"
    yellow: str = "\x1b[33m"
    red: str = "\x1b[31m"
    cyan: str = "\x1b[1;36m"
    bold_red: str = "\x1b[31;1m"
    purple: str = "\x1b[38;5;105m"
    blue: str = "\x1b[34m"
    reset: str = "\x1b[0m"

    def _get_format_template(self, record: logging.LogRecord) -> str:
        """Get the format template based on the logging level.

        Args:
            record: The log record to format.

        Returns:
            The format template string.
        """
        if get_logging_level() == LoggingLevels.DEBUG:
            return "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
        else:
            return "%(message)s"

    COLORS: Dict[LoggingLevels, str] = {
        LoggingLevels.DEBUG: grey,
        LoggingLevels.INFO: white,
        LoggingLevels.WARN: yellow,
        LoggingLevels.ERROR: red,
        LoggingLevels.CRITICAL: bold_red,
    }

    def format(self, record: logging.LogRecord) -> str:
        """Converts a log record to a (colored) string.

        Args:
            record: LogRecord generated by the code.

        Returns:
            A string formatted according to specifications.
        """
        format_template = self._get_format_template(record)

        message = record.getMessage()
        try:
            if step_names_in_console.get():
                message = _add_step_name_to_message(message)
        except Exception:
            pass

        modified_record = logging.LogRecord(
            name=record.name,
            level=record.levelno,
            pathname=record.pathname,
            lineno=record.lineno,
            msg=message,
            args=(),
            exc_info=record.exc_info,
        )

        if ZENML_LOGGING_COLORS_DISABLED:
            formatter = logging.Formatter(format_template)
            return formatter.format(modified_record)
        else:
            log_fmt = (
                self.COLORS[LoggingLevels(record.levelno)]
                + format_template
                + self.reset
            )
            formatter = logging.Formatter(log_fmt)
            formatted_message = formatter.format(modified_record)
            quoted_groups = re.findall("`([^`]*)`", formatted_message)
            for quoted in quoted_groups:
                formatted_message = formatted_message.replace(
                    "`" + quoted + "`",
                    self.reset
                    + self.purple
                    + quoted
                    + self.COLORS.get(LoggingLevels(record.levelno)),
                )

            url_pattern = r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+"
            urls = re.findall(url_pattern, formatted_message)
            for url in urls:
                formatted_message = formatted_message.replace(
                    url,
                    self.reset
                    + self.blue
                    + url
                    + self.COLORS.get(LoggingLevels(record.levelno)),
                )
            return formatted_message


def get_logging_level() -> LoggingLevels:
    """Get logging level from the env variable.

    Returns:
        The logging level.

    Raises:
        KeyError: If the logging level is not found.
    """
    verbosity = ZENML_LOGGING_VERBOSITY.upper()
    if verbosity not in LoggingLevels.__members__:
        raise KeyError(
            f"Verbosity must be one of {list(LoggingLevels.__members__.keys())}"
        )

    if ZENML_STORAGE_LOGGING_VERBOSITY is not None:
        get_logger(__name__).warning(
            "The ZENML_STORAGE_LOGGING_VERBOSITY is no longer supported. "
            "Please use the ZENML_LOGGING_VERBOSITY instead."
        )

    return LoggingLevels[verbosity]


def set_root_verbosity() -> None:
    """Set the root verbosity."""
    level = get_logging_level()
    if level != LoggingLevels.NOTSET:
        if ENABLE_RICH_TRACEBACK:
            rich_tb_install(show_locals=(level == LoggingLevels.DEBUG))

        logging.root.setLevel(level=level.value)
        get_logger(__name__).debug(
            f"Logging set to level: {logging.getLevelName(level.value)}"
        )
    else:
        logging.disable(sys.maxsize)
        logging.getLogger().disabled = True
        get_logger(__name__).debug("Logging NOTSET")


def _wrapped_write(original_write: Any, stream_name: str) -> Any:
    """Wrap stdout/stderr write method to route logs to LoggingContext.

    Args:
        original_write: The original write method.
        stream_name: The name of the stream.

    Returns:
        The wrapped write method.
    """

    def wrapped_write(text: str) -> Any:
        """Write method that routes logs through LoggingContext.

        Args:
            text: The text to write.

        Returns:
            The result of the original write method.
        """
        from zenml.utils.logging_utils import LoggingContext

        level = logging.INFO if stream_name == "stdout" else logging.ERROR

        if logging.root.isEnabledFor(level):
            record = logging.LogRecord(
                name=stream_name,
                level=level,
                pathname="",
                lineno=0,
                msg=text,
                args=(),
                exc_info=None,
                func="",
            )
            LoggingContext.emit(record)

        return original_write(text)

    return wrapped_write


def wrap_stdout_stderr() -> None:
    """Wrap stdout and stderr write methods to route through LoggingContext."""
    global _stdout_wrapped, _stderr_wrapped
    global _original_stdout_write, _original_stderr_write

    if not _stdout_wrapped:
        _original_stdout_write = getattr(sys.stdout, "write")
        setattr(
            sys.stdout,
            "write",
            _wrapped_write(_original_stdout_write, "stdout"),
        )
        _stdout_wrapped = True

    if not _stderr_wrapped:
        _original_stderr_write = getattr(sys.stderr, "write")
        setattr(
            sys.stderr,
            "write",
            _wrapped_write(_original_stderr_write, "stderr"),
        )
        _stderr_wrapped = True


class ZenMLLoggingHandler(logging.Handler):
    """Custom handler that routes logs through LoggingContext."""

    def emit(self, record: logging.LogRecord) -> None:
        """Emit a log record through LoggingContext.

        Args:
            record: The log record to emit.
        """
        from zenml.utils.logging_utils import LoggingContext

        LoggingContext.emit(record)


def get_console_handler() -> logging.Handler:
    """Get console handler that writes to original stdout.

    Returns:
        A console handler.
    """
    handler = logging.StreamHandler(_ZenMLStdoutStream())
    handler.setFormatter(ConsoleFormatter())
    return handler


def get_zenml_handler() -> logging.Handler:
    """Get ZenML handler that routes logs through LoggingContext.

    Returns:
        A ZenML handler.
    """
    return ZenMLLoggingHandler()


def init_logging() -> None:
    """Initialize the logging system."""
    set_root_verbosity()
    wrap_stdout_stderr()

    # Add both handlers to the root logger
    root_logger = logging.getLogger()

    # Console handler - writes to original stdout
    root_logger.addHandler(get_console_handler())

    # ZenML handler - routes through LoggingContext
    root_logger.addHandler(get_zenml_handler())

    # Mute tensorflow cuda warnings
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

    # logging capture warnings
    logging.captureWarnings(True)

    # Enable logs if environment variable SUPPRESS_ZENML_LOGS is not set to True
    suppress_zenml_logs: bool = handle_bool_env_var(
        ENV_ZENML_SUPPRESS_LOGS, True
    )
    if suppress_zenml_logs:
        # suppress logger info messages
        suppressed_logger_names = [
            "urllib3",
            "azure.core.pipeline.policies.http_logging_policy",
            "grpc",
            "requests",
            "kfp",
            "tensorflow",
        ]
        for logger_name in suppressed_logger_names:
            logging.getLogger(logger_name).setLevel(logging.WARNING)

        # disable logger messages
        disabled_logger_names = [
            "rdbms_metadata_access_object",
            "backoff",
            "segment",
        ]
        for logger_name in disabled_logger_names:
            logging.getLogger(logger_name).setLevel(logging.WARNING)
            logging.getLogger(logger_name).disabled = True
